import unittest

from mock import Mock, sentinel, patch, MagicMock

from morelia.matchers import (MethodNameStepMatcher, IStepMatcher,
                              RegexpStepMatcher, ParseStepMatcher)


class TestStepMatcher(IStepMatcher):

    def match(self, predicate, augmented_predicate, step_methods):
        pass  # pragma: nocover

    def suggest(self, predicate):
        pass  # pragma: nocover


class IStepMatcherAddMatcherTestCase(unittest.TestCase):
    """ Test :py:meth:`IStepMatcher.add_matcher`. """

    def test_should_add_matcher(self):
        """ Scenario: add matcher """
        # Arrange
        suite = Mock()
        matcher1 = TestStepMatcher(suite)
        matcher2 = TestStepMatcher(suite)
        # Act
        matcher1.add_matcher(matcher2)
        # Assert
        self.assertEqual(matcher1._next, matcher2)

    def test_should_delegate_adding_matcher(self):
        """ Scenario: delegating add """
        # Arrange
        suite = Mock()
        matcher1 = TestStepMatcher(suite)
        matcher2 = TestStepMatcher(suite)
        matcher3 = TestStepMatcher(suite)
        matcher1._next = matcher2
        # Act
        matcher1.add_matcher(matcher3)
        # Assert
        self.assertEqual(matcher1._next, matcher2)
        self.assertEqual(matcher2._next, matcher3)

    def test_should_chain_adding_matchers(self):
        """ Scenario: chaining """
        # Arrange
        suite = Mock()
        matcher1 = TestStepMatcher(suite)
        matcher2 = TestStepMatcher(suite)
        matcher3 = TestStepMatcher(suite)
        # Act
        matcher1.add_matcher(matcher2).add_matcher(matcher3)
        # Assert
        self.assertEqual(matcher1._next, matcher2)
        self.assertEqual(matcher2._next, matcher3)


class IStepMatcherFindTestCase(unittest.TestCase):
    """ Test :py:meth:`IStepMatcher.find`. """

    def test_should_find_method_when_step_methods_given(self):
        """ Scenario: find method when step methods given"""
        # Arrange
        suite = Mock()
        obj = TestStepMatcher(suite)
        # Act
        step_methods = sentinel.step_methods
        predicate = sentinel.predicate
        augmented_predicate = sentinel.augmented_predicate
        with patch.object(obj, 'match') as match:
            match.return_value = (sentinel.method, (), {})
            method, args, kwargs = obj.find(predicate, augmented_predicate, step_methods)
            match.assert_called_once_with(predicate, augmented_predicate, step_methods)
        # Assert
        self.assertEqual(method, sentinel.method)

    def test_should_find_method_when_no_step_methods_given(self):
        """ Scenario: no step methods """
        # Arrange
        suite = Mock()
        obj = TestStepMatcher(suite)
        # Act
        predicate = sentinel.predicate
        augmented_predicate = sentinel.augmented_predicate
        with patch.object(obj, 'match') as match:
            with patch.object(obj, '_get_all_step_methods') as _get_all_step_methods:
                _get_all_step_methods.return_value = sentinel.steps
                match.return_value = (sentinel.method, (), {})
                method, args, kwargs = obj.find(predicate, augmented_predicate)
                match.assert_called_once_with(predicate, augmented_predicate, sentinel.steps)
        # Assert
        self.assertEqual(method, sentinel.method)

    def test_should_not_find_method(self):
        """ Scenario: not found """
        # Arrange
        suite = Mock()
        obj = TestStepMatcher(suite)
        # Act
        step_methods = sentinel.step_methods
        predicate = sentinel.predicate
        augmented_predicate = sentinel.augmented_predicate
        with patch.object(obj, 'match') as match:
            match.return_value = (None, (), {})
            method, args, kwargs = obj.find(predicate, augmented_predicate, step_methods)
            match.assert_called_once_with(predicate, augmented_predicate, step_methods)
        # Assert
        self.assertEqual(method, None)

    def test_should_delegate_search(self):
        """ Scenario: delegate """
        # Arrange
        suite = Mock()
        matcher1 = TestStepMatcher(suite)
        matcher2 = TestStepMatcher(suite)
        matcher1.add_matcher(matcher2)
        # Act
        step_methods = sentinel.step_methods
        predicate = sentinel.predicate
        augmented_predicate = sentinel.augmented_predicate
        with patch.object(matcher1, 'match') as match:
            match.return_value = (None, (), {})
            with patch.object(matcher2, 'find') as find:
                find.return_value = (sentinel.method, (), {})
                method, args, kwargs = matcher1.find(predicate, augmented_predicate, step_methods)
                match.assert_called_once_with(predicate, augmented_predicate, step_methods)
                # Assert
                find.assert_called_once_with(predicate, augmented_predicate, step_methods)
                self.assertEqual(method, sentinel.method)


class IStepMatcherGetAllStepMethodsTestCase(unittest.TestCase):
    """ Test :py:meth:`IStepMatcher._get_all_step_methods`. """

    def test_should_return_steps_list(self):
        """ Scenario: filtered step lists """
        # Arrange
        step_methods = ['step_method1', 'step_method2']
        all_methods = step_methods + ['method3']
        suite = MagicMock()
        suite.__dir__ = Mock(return_value=all_methods)
        obj = TestStepMatcher(suite)
        # Act
        result = obj._get_all_step_methods()
        # Assert
        self.assertEqual(result, step_methods)


class MethodNameStepMatcherMatchTestCase(unittest.TestCase):
    """ Test :py:meth:`MethodNameStepMatcher.match`. """

    def test_should_return_method(self):
        """ Scenario: match by name """
        # Arrange
        predicate = 'my_milkshake'
        augmented_predicate = 'my_milkshake'
        method_name = 'step_%s' % predicate
        methods = {
            method_name: sentinel.method
        }
        suite = Mock(**methods)
        obj = MethodNameStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, sentinel.method)

    def test_should_return_none_if_method_not_found(self):
        """ Scenario: no method """
        # Arrange
        predicate = 'not there'
        augmented_predicate = 'my_milkshake'
        method_name = 'step_%s' % predicate
        methods = {
            method_name: sentinel.method
        }
        suite = Mock(**methods)
        obj = MethodNameStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, None)

    def test_should_return_none_if_method_name_too_short(self):
        """ Scenario: method too short """
        # Arrange
        predicate = 'my milk'
        augmented_predicate = 'my_milkshake'
        method_name = 'step_%s' % predicate
        methods = {
            method_name: sentinel.method
        }
        suite = Mock(**methods)
        obj = MethodNameStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, None)


class MethodNameStepMatcherSuggestTestCase(unittest.TestCase):
    """ Test :py:meth:`MethodNameStepMatcher.suggest`. """

    def test_should_return_suggested_method(self):
        """ Scenariusz: suggest """
        # Arrange
        obj = MethodNameStepMatcher(sentinel.suite)
        pattern = u'    def step_%(method_name)s(self%(args)s):\n\n        # code\n        pass\n\n'
        test_data = [
            ('tastes great', 'tastes_great', ''),
            ('less filling', 'less_filling', ''),
            ('line\nfeed', 'line_feed', ''),
            ('tick\'ed\'', 'tick_ed', ''),
            ('tastes   great', 'tastes_great', ''),
            ('argu<ment>al', 'argu_ment_al', ', ment'),
            ('arg<u>ment<al>', 'arg_u_ment_al', ', u, al'),
            ('str"ing"', 'str_ing', ', ing'),
            ('"str"i"ngs"', 'str_i_ngs', ', str, ngs'),
        ]
        for predicate, method, args in test_data:
            # Act
            result = obj.suggest(predicate)
            # Assert
            expected = pattern % {
                'method_name': method,
                'args': args,
            }
            self.assertEqual(result, expected)


class DocStringStepMatcherMatchTestCase(unittest.TestCase):
    """ Test :py:meth:`DocStringStepMatcher.match`. """

    def test_should_return_method_and_args(self):
        """ Scenario: match by docstring """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the (boys|girls) to (.*) yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = RegexpStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_args, ('boys', 'the'))

    def test_should_return_method_and_kwargs(self):
        """ Scenario: match by docstring with named groups """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the (?P<who>boys|girls) to (?P<other>.*) yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = RegexpStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_kwargs, {'who': 'boys', 'other': 'the'})

    def test_should_return_method_and_kwargs_with_mixed_groups(self):
        """ Scenario: match by docstring with named and not named groups """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the (?P<who>boys|girls) to (.*) yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = RegexpStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_args, ())
        self.assertEqual(result_kwargs, {'who': 'boys'})

    def test_should_return_none_if_docstring_not_mached(self):
        """ Scenario: no match by docstring """
        # Arrange
        predicate = 'not there'
        augmented_predicate = 'not there'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the (boys|girls) to (.*) yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = RegexpStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertTrue(result_method is None)
        self.assertEqual(result_args, ())

    def test_should_return_none_if_no_docstring(self):
        """ Scenario: no match by docstring """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        method = Mock(__doc__='')
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = RegexpStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertTrue(result_method is None)
        self.assertEqual(result_args, ())

    def test_should_return_second_method_and_matches(self):
        """ Scenario: many methods """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the (boys|girls) to (.*) yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method,
            'step_other': sentinel.method,
        }
        suite = Mock(**methods)
        obj = RegexpStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(
            predicate, augmented_predicate, [method_name, 'step_other'])
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_args, ('boys', 'the'))


class DocStringStepMatcherSuggestTestCase(unittest.TestCase):
    """ Test :py:meth:`DocStringStepMatcher.suggest`. """

    def test_should_return_suggested_method(self):
        """ Scenariusz: suggest """
        # Arrange
        obj = RegexpStepMatcher(sentinel.suite)
        # Act
        pattern = u'    def step_%(method_name)s(self%(args)s):\n        %(docstring)s\n\n        # code\n        pass\n\n'
        test_data = [
            ('tastes great', 'tastes_great', r"ur'tastes great'", ''),
            ('less filling', 'less_filling', r"ur'less filling'", ''),
            ('line\nfeed', 'line_feed', r"ur'line\nfeed'", ''),
            ('tick\'ed\'', 'tick_ed', r"ur'tick\'ed\''", ''),
            ('tastes   great', 'tastes_great', r"ur'tastes\s+great'", ''),
            ('argu<ment>al', 'argu_ment_al', r"ur'argu(.+)al'", ', ment'),
            ('arg<u>ment<al>', 'arg_u_ment_al', r"ur'arg(.+)ment(.+)'", ', u, al'),
            ('str"ing"', 'str_ing', 'ur\'str"([^"]+)"\'', ', ing'),
            ('"str"i"ngs"', 'str_i_ngs', 'ur\'"([^"]+)"i"([^"]+)"\'', ', str, ngs'),
        ]
        for predicate, method, docstring, args in test_data:
            result = obj.suggest(predicate)
            # Assert
            expected = pattern % {
                'method_name': method,
                'docstring': docstring,
                'args': args,
            }
            self.assertEqual(result, expected)


class ParseStepMatcherMatchTestCase(unittest.TestCase):
    """ Test :py:meth:`ParseStepMatcher.match`. """

    def test_should_return_method_and_args(self):
        """ Scenario: match by docstring """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the {} to {} yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = ParseStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_args, ('boys', 'the'))

    def test_should_return_method_and_kwargs(self):
        """ Scenario: match by docstring with named groups """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the {who} to {other} yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = ParseStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_kwargs, {'who': 'boys', 'other': 'the'})

    def test_should_return_method_args_and_kwargs_with_mixed_groups(self):
        """ Scenario: match by docstring with named and not named groups """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the {who} to {} yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = ParseStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_args, ('the',))
        self.assertEqual(result_kwargs, {'who': 'boys'})

    def test_should_return_none_if_docstring_not_mached(self):
        """ Scenario: no match by docstring """
        # Arrange
        predicate = 'not there'
        augmented_predicate = 'not there'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the {who} to {} yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = ParseStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertTrue(result_method is None)
        self.assertEqual(result_args, ())

    def test_should_return_none_if_no_docstring(self):
        """ Scenario: no match by docstring """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        method = Mock(__doc__='')
        methods = {
            method_name: method
        }
        suite = Mock(**methods)
        obj = ParseStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(predicate, augmented_predicate, methods.keys())
        # Assert
        self.assertTrue(result_method is None)
        self.assertEqual(result_args, ())

    def test_should_return_second_method_and_matches(self):
        """ Scenario: many methods """
        # Arrange
        predicate = 'my milkshake brings all the boys to the yard'
        augmented_predicate = 'my milkshake brings all the boys to the yard'
        method_name = 'step_%s' % predicate
        docstring = r'my milkshake brings all the {who} to {} yard'
        method = Mock(__doc__=docstring)
        methods = {
            method_name: method,
            'step_other': sentinel.method,
        }
        suite = Mock(**methods)
        obj = ParseStepMatcher(suite)
        # Act
        result_method, result_args, result_kwargs = obj.match(
            predicate, augmented_predicate, [method_name, 'step_other'])
        # Assert
        self.assertEqual(result_method, method)
        self.assertEqual(result_args, ('the',))
        self.assertEqual(result_kwargs, {'who': 'boys'})


class ParseStepMatcherSuggestTestCase(unittest.TestCase):
    """ Test :py:meth:`ParseStepMatcher.suggest`. """

    def test_should_return_suggested_method(self):
        """ Scenario: suggest """
        # Arrange
        obj = ParseStepMatcher(sentinel.suite)
        # Act
        pattern = u'    def step_%(method_name)s(self%(args)s):\n        %(docstring)s\n\n        # code\n        pass\n\n'
        test_data = [
            ('tastes great', 'tastes_great', r"ur'tastes great'", ''),
            ('less filling', 'less_filling', r"ur'less filling'", ''),
            ('line\nfeed', 'line_feed', r"ur'line\nfeed'", ''),
            ('tick\'ed\'', 'tick_ed', r"ur'tick\'ed\''", ''),
            ('tastes   great', 'tastes_great', r"ur'tastes\s+great'", ''),
            ('argu<ment>al', 'argu_ment_al', r"ur'argu{ment}al'", ', ment'),
            ('arg<u>ment<al>', 'arg_u_ment_al', r"ur'arg{u}ment{al}'", ', u, al'),
            ('str"ing"', 'str_ing', 'ur\'str"{ing}"\'', ', ing'),
            ('"str"i"ngs"', 'str_i_ngs', 'ur\'"{str}"i"{ngs}"\'', ', str, ngs'),
        ]
        for predicate, method, docstring, args in test_data:
            result = obj.suggest(predicate)
            # Assert
            expected = pattern % {
                'method_name': method,
                'docstring': docstring,
                'args': args,
            }
            self.assertEqual(result, expected)



# -*- coding: utf-8 -*-

# source: https://github.com/cucumber/gherkin/blob/master/lib/gherkin/i18n.json

TRANSLATIONS = {
    "en": {
        "name": u"English",
        "native": u"English",
        "feature": u"Feature|Business Need|Ability",
        "background": u"Background",
        "scenario": u"Scenario",
        "scenario_outline": u"Scenario Outline|Scenario Template",
        "examples": u"Examples|Scenarios",
        "given": u"Given",
        "when": u"When",
        "then": u"Then",
        "and": u"And",
        "but": u"But"
    },
    "af": {
        "name": u"Afrikaans",
        "native": u"Afrikaans",
        "feature": u"Funksie|Besigheid Behoefte|Vermoë",
        "background": u"Agtergrond",
        "scenario": u"Situasie",
        "scenario_outline": u"Situasie Uiteensetting",
        "examples": u"Voorbeelde",
        "given": u"Gegewe",
        "when": u"Wanneer",
        "then": u"Dan",
        "and": u"En",
        "but": u"Maar"
    },
    "ar": {
        "name": u"Arabic",
        "native": u"العربية",
        "feature": u"خاصية",
        "background": u"الخلفية",
        "scenario": u"سيناريو",
        "scenario_outline": u"سيناريو مخطط",
        "examples": u"امثلة",
        "given": u"بفرض",
        "when": u"متى|عندما",
        "then": u"اذاً|ثم",
        "and": u"و",
        "but": u"لكن"
    },
    "bm": {
        "name": u"Malay",
        "native": u"Bahasa Melayu",
        "feature": u"Fungsi",
        "background": u"Latar Belakang",
        "scenario": u"Senario|Situai|Keadaan",
        "scenario_outline": u"Template Senario|Template Situai|Template Keadaan|Menggariskan Senario",
        "examples": u"Contoh",
        "given": u"Diberi|Bagi",
        "when": u"Apabila",
        "then": u"Maka|Kemudian",
        "and": u"Dan",
        "but": u"Tetapi|Tapi"
    },
    "bg": {
        "name": u"Bulgarian",
        "native": u"български",
        "feature": u"Функционалност",
        "background": u"Предистория",
        "scenario": u"Сценарий",
        "scenario_outline": u"Рамка на сценарий",
        "examples": u"Примери",
        "given": u"Дадено",
        "when": u"Когато",
        "then": u"То",
        "and": u"И",
        "but": u"Но"
    },
    "ca": {
        "name": u"Catalan",
        "native": u"català",
        "background": u"Rerefons|Antecedents",
        "feature": u"Característica|Funcionalitat",
        "scenario": u"Escenari",
        "scenario_outline": u"Esquema de l'escenari",
        "examples": u"Exemples",
        "given": u"Donat|Donada|Atès|Atesa",
        "when": u"Quan",
        "then": u"Aleshores|Cal",
        "and": u"I",
        "but": u"Però"
    },
    "cy-GB": {
        "name": u"Welsh",
        "native": u"Cymraeg",
        "background": u"Cefndir",
        "feature": u"Arwedd",
        "scenario": u"Scenario",
        "scenario_outline": u"Scenario Amlinellol",
        "examples": u"Enghreifftiau",
        "given": u"Anrhegedig a",
        "when": u"Pryd",
        "then": u"Yna",
        "and": u"A",
        "but": u"Ond"
    },
    "cs": {
        "name": u"Czech",
        "native": u"Česky",
        "feature": u"Požadavek",
        "background": u"Pozadí|Kontext",
        "scenario": u"Scénář",
        "scenario_outline": u"Náčrt Scénáře|Osnova scénáře",
        "examples": u"Příklady",
        "given": u"Pokud|Za předpokladu",
        "when": u"Když",
        "then": u"Pak",
        "and": u"A také|A",
        "but": u"Ale"
    },
    "da": {
        "name": u"Danish",
        "native": u"dansk",
        "feature": u"Egenskab",
        "background": u"Baggrund",
        "scenario": u"Scenarie",
        "scenario_outline": u"Abstrakt Scenario",
        "examples": u"Eksempler",
        "given": u"Givet",
        "when": u"Når",
        "then": u"Så",
        "and": u"Og",
        "but": u"Men"
    },
    "de": {
        "name": u"German",
        "native": u"Deutsch",
        "feature": u"Funktionalität",
        "background": u"Grundlage",
        "scenario": u"Szenario",
        "scenario_outline": u"Szenariogrundriss",
        "examples": u"Beispiele",
        "given": u"Angenommen|Gegeben sei|Gegeben seien",
        "when": u"Wenn",
        "then": u"Dann",
        "and": u"Und",
        "but": u"Aber"
    },
    "el": {
        "name": u"Greek",
        "native": u"Ελληνικά",
        "feature": u"Δυνατότητα|Λειτουργία",
        "background": u"Υπόβαθρο",
        "scenario": u"Σενάριο",
        "scenario_outline": u"Περιγραφή Σεναρίου",
        "examples": u"Παραδείγματα|Σενάρια",
        "given": u"Δεδομένου",
        "when": u"Όταν",
        "then": u"Τότε",
        "and": u"Και",
        "but": u"Αλλά"
    },
    "en-au": {
        "name": u"Australian",
        "native": u"Australian",
        "feature": u"Pretty much",
        "background": u"First off",
        "scenario": u"Awww, look mate",
        "scenario_outline": u"Reckon it's like",
        "examples": u"You'll wanna",
        "given": u"Y'know",
        "when": u"It's just unbelievable",
        "then": u"But at the end of the day I reckon",
        "and": u"Too right",
        "but": u"Yeah nah"
    },
    "en-lol": {
        "name": u"LOLCAT",
        "native": u"LOLCAT",
        "feature": u"OH HAI",
        "background": u"B4",
        "scenario": u"MISHUN",
        "scenario_outline": u"MISHUN SRSLY",
        "examples": u"EXAMPLZ",
        "given": u"I CAN HAZ",
        "when": u"WEN",
        "then": u"DEN",
        "and": u"AN",
        "but": u"BUT"
    },
    "en-old": {
        "name": u"Old English",
        "native": u"Englisc",
        "feature": u"Hwaet|Hwæt",
        "background": u"Aer|Ær",
        "scenario": u"Swa",
        "scenario_outline": u"Swa hwaer swa|Swa hwær swa",
        "examples": u"Se the|Se þe|Se ðe",
        "given": u"Thurh|Þurh|Ðurh",
        "when": u"Tha|Þa|Ða",
        "then": u"Tha|Þa|Ða|Tha the|Þa þe|Ða ðe",
        "and": u"Ond|7",
        "but": u"Ac"
    },
    "en-pirate": {
        "name": u"Pirate",
        "native": u"Pirate",
        "feature": u"Ahoy matey!",
        "background": u"Yo-ho-ho",
        "scenario": u"Heave to",
        "scenario_outline": u"Shiver me timbers",
        "examples": u"Dead men tell no tales",
        "given": u"Gangway!",
        "when": u"Blimey!",
        "then": u"Let go and haul",
        "and": u"Aye",
        "but": u"Avast!"
    },
    "en-Scouse": {
        "name": u"Scouse",
        "native": u"Scouse",
        "feature": u"Feature",
        "background": u"Dis is what went down",
        "scenario": u"The thing of it is",
        "scenario_outline": u"Wharrimean is",
        "examples": u"Examples",
        "given": u"Givun|Youse know when youse got",
        "when": u"Wun|Youse know like when",
        "then": u"Dun|Den youse gotta",
        "and": u"An",
        "but": u"Buh"
    },
    "en-tx": {
        "name": u"Texan",
        "native": u"Texan",
        "feature": u"Feature",
        "background": u"Background",
        "scenario": u"Scenario",
        "scenario_outline": u"All y'all",
        "examples": u"Examples",
        "given": u"Given y'all",
        "when": u"When y'all",
        "then": u"Then y'all",
        "and": u"And y'all",
        "but": u"But y'all"
    },
    "eo": {
        "name": u"Esperanto",
        "native": u"Esperanto",
        "feature": u"Trajto",
        "background": u"Fono",
        "scenario": u"Scenaro",
        "scenario_outline": u"Konturo de la scenaro",
        "examples": u"Ekzemploj",
        "given": u"Donitaĵo",
        "when": u"Se",
        "then": u"Do",
        "and": u"Kaj",
        "but": u"Sed"
    },
    "es": {
        "name": u"Spanish",
        "native": u"español",
        "background": u"Antecedentes",
        "feature": u"Característica",
        "scenario": u"Escenario",
        "scenario_outline": u"Esquema del escenario",
        "examples": u"Ejemplos",
        "given": u"Dado|Dada|Dados|Dadas",
        "when": u"Cuando",
        "then": u"Entonces",
        "and": u"Y",
        "but": u"Pero"
    },
    "et": {
        "name": u"Estonian",
        "native": u"eesti keel",
        "feature": u"Omadus",
        "background": u"Taust",
        "scenario": u"Stsenaarium",
        "scenario_outline": u"Raamstsenaarium",
        "examples": u"Juhtumid",
        "given": u"Eeldades",
        "when": u"Kui",
        "then": u"Siis",
        "and": u"Ja",
        "but": u"Kuid"
    },
    "fa": {
        "name": u"Persian",
        "native": u"فارسی",
        "feature": u"وِیژگی",
        "background": u"زمینه",
        "scenario": u"سناریو",
        "scenario_outline": u"الگوی سناریو",
        "examples": u"نمونه ها",
        "given": u"با فرض",
        "when": u"هنگامی",
        "then": u"آنگاه",
        "and": u"و",
        "but": u"اما"
    },
    "fi": {
        "name": u"Finnish",
        "native": u"suomi",
        "feature": u"Ominaisuus",
        "background": u"Tausta",
        "scenario": u"Tapaus",
        "scenario_outline": u"Tapausaihio",
        "examples": u"Tapaukset",
        "given": u"Oletetaan",
        "when": u"Kun",
        "then": u"Niin",
        "and": u"Ja",
        "but": u"Mutta"
    },
    "fr": {
        "name": u"French",
        "native": u"français",
        "feature": u"Fonctionnalité",
        "background": u"Contexte",
        "scenario": u"Scénario",
        "scenario_outline": u"Plan du scénario|Plan du Scénario",
        "examples": u"Exemples",
        "given": u"Soit|Etant donné|Etant donnée|Etant donnés|Etant données|Étant donné|Étant donnée|Étant donnés|Étant données",
        "when": u"Quand|Lorsque|Lorsqu'<",
        "then": u"Alors",
        "and": u"Et",
        "but": u"Mais"
    },
    "gl": {
        "name": u"Galician",
        "native": u"galego",
        "background": u"Contexto",
        "feature": u"Característica",
        "scenario": u"Escenario",
        "scenario_outline": u"Esbozo do escenario",
        "examples": u"Exemplos",
        "given": u"Dado|Dada|Dados|Dadas",
        "when": u"Cando",
        "then": u"Entón|Logo",
        "and": u"E",
        "but": u"Mais|Pero"
    },
    "he": {
        "name": u"Hebrew",
        "native": u"עברית",
        "feature": u"תכונה",
        "background": u"רקע",
        "scenario": u"תרחיש",
        "scenario_outline": u"תבנית תרחיש",
        "examples": u"דוגמאות",
        "given": u"בהינתן",
        "when": u"כאשר",
        "then": u"אז|אזי",
        "and": u"וגם",
        "but": u"אבל"
    },
    "hi": {
        "name": u"Hindi",
        "native": u"हिंदी",
        "feature": u"रूप लेख",
        "background": u"पृष्ठभूमि",
        "scenario": u"परिदृश्य",
        "scenario_outline": u"परिदृश्य रूपरेखा",
        "examples": u"उदाहरण",
        "given": u"अगर|यदि|चूंकि",
        "when": u"जब|कदा",
        "then": u"तब|तदा",
        "and": u"और|तथा",
        "but": u"पर|परन्तु|किन्तु"
    },
    "hr": {
        "name": u"Croatian",
        "native": u"hrvatski",
        "feature": u"Osobina|Mogućnost|Mogucnost",
        "background": u"Pozadina",
        "scenario": u"Scenarij",
        "scenario_outline": u"Skica|Koncept",
        "examples": u"Primjeri|Scenariji",
        "given": u"Zadan|Zadani|Zadano",
        "when": u"Kada|Kad",
        "then": u"Onda",
        "and": u"I",
        "but": u"Ali"
    },
    "ht": {
        "name": u"Creole",
        "native": u"kreyòl",
        "feature": u"Karakteristik|Mak|Fonksyonalite",
        "background": u"Kontèks|Istorik",
        "scenario": u"Senaryo",
        "scenario_outline": u"Plan senaryo|Plan Senaryo|Senaryo deskripsyon|Senaryo Deskripsyon|Dyagram senaryo|Dyagram Senaryo",
        "examples": u"Egzanp",
        "given": u"Sipoze|Sipoze ke|Sipoze Ke",
        "when": u"Lè|Le",
        "then": u"Lè sa a|Le sa a",
        "and": u"Ak|Epi|E",
        "but": u"Men"
    },
    "hu": {
        "name": u"Hungarian",
        "native": u"magyar",
        "feature": u"Jellemző",
        "background": u"Háttér",
        "scenario": u"Forgatókönyv",
        "scenario_outline": u"Forgatókönyv vázlat",
        "examples": u"Példák",
        "given": u"Amennyiben|Adott",
        "when": u"Majd|Ha|Amikor",
        "then": u"Akkor",
        "and": u"És",
        "but": u"De"
    },
    "id": {
        "name": u"Indonesian",
        "native": u"Bahasa Indonesia",
        "feature": u"Fitur",
        "background": u"Dasar",
        "scenario": u"Skenario",
        "scenario_outline": u"Skenario konsep",
        "examples": u"Contoh",
        "given": u"Dengan",
        "when": u"Ketika",
        "then": u"Maka",
        "and": u"Dan",
        "but": u"Tapi"
    },
    "is": {
        "name": u"Icelandic",
        "native": u"Íslenska",
        "feature": u"Eiginleiki",
        "background": u"Bakgrunnur",
        "scenario": u"Atburðarás",
        "scenario_outline": u"Lýsing Atburðarásar|Lýsing Dæma",
        "examples": u"Dæmi|Atburðarásir",
        "given": u"Ef",
        "when": u"Þegar",
        "then": u"Þá",
        "and": u"Og",
        "but": u"En"
    },
    "it": {
        "name": u"Italian",
        "native": u"italiano",
        "feature": u"Funzionalità",
        "background": u"Contesto",
        "scenario": u"Scenario",
        "scenario_outline": u"Schema dello scenario",
        "examples": u"Esempi",
        "given": u"Dato|Data|Dati|Date",
        "when": u"Quando",
        "then": u"Allora",
        "and": u"E",
        "but": u"Ma"
    },
    "ja": {
        "name": u"Japanese",
        "native": u"日本語",
        "feature": u"フィーチャ|機能",
        "background": u"背景",
        "scenario": u"シナリオ",
        "scenario_outline": u"シナリオアウトライン|シナリオテンプレート|テンプレ|シナリオテンプレ",
        "examples": u"例|サンプル",
        "given": u"前提<",
        "when": u"もし<",
        "then": u"ならば<",
        "and": u"かつ<",
        "but": u"しかし<|但し<|ただし<"
    },
    "jv": {
        "name": u"Javanese",
        "native": u"Basa Jawa",
        "feature": u"Fitur",
        "background": u"Dasar",
        "scenario": u"Skenario",
        "scenario_outline": u"Konsep skenario",
        "examples": u"Conto|Contone",
        "given": u"Nalika|Nalikaning",
        "when": u"Manawa|Menawa",
        "then": u"Njuk|Banjur",
        "and": u"Lan",
        "but": u"Tapi|Nanging|Ananging"
    },
    "kn": {
        "name": u"Kannada",
        "native": u"ಕನ್ನಡ",
        "background": u"ಹಿನ್ನೆಲೆ",
        "feature": u"ಹೆಚ್ಚಳ",
        "scenario": u"ಕಥಾಸಾರಾಂಶ",
        "scenario_outline": u"ವಿವರಣೆ",
        "examples": u"ಉದಾಹರಣೆಗಳು",
        "given": u"ನೀಡಿದ",
        "when": u"ಸ್ಥಿತಿಯನ್ನು",
        "then": u"ನಂತರ",
        "and": u"ಮತ್ತು",
        "but": u"ಆದರೆ"
    },
    "ko": {
        "name": u"Korean",
        "native": u"한국어",
        "background": u"배경",
        "feature": u"기능",
        "scenario": u"시나리오",
        "scenario_outline": u"시나리오 개요",
        "examples": u"예",
        "given": u"조건<|먼저<",
        "when": u"만일<|만약<",
        "then": u"그러면<",
        "and": u"그리고<",
        "but": u"하지만<|단<"
    },
    "lt": {
        "name": u"Lithuanian",
        "native": u"lietuvių kalba",
        "feature": u"Savybė",
        "background": u"Kontekstas",
        "scenario": u"Scenarijus",
        "scenario_outline": u"Scenarijaus šablonas",
        "examples": u"Pavyzdžiai|Scenarijai|Variantai",
        "given": u"Duota",
        "when": u"Kai",
        "then": u"Tada",
        "and": u"Ir",
        "but": u"Bet"
    },
    "lu": {
        "name": u"Luxemburgish",
        "native": u"Lëtzebuergesch",
        "feature": u"Funktionalitéit",
        "background": u"Hannergrond",
        "scenario": u"Szenario",
        "scenario_outline": u"Plang vum Szenario",
        "examples": u"Beispiller",
        "given": u"ugeholl",
        "when": u"wann",
        "then": u"dann",
        "and": u"an|a",
        "but": u"awer|mä"
    },
    "lv": {
        "name": u"Latvian",
        "native": u"latviešu",
        "feature": u"Funkcionalitāte|Fīča",
        "background": u"Konteksts|Situācija",
        "scenario": u"Scenārijs",
        "scenario_outline": u"Scenārijs pēc parauga",
        "examples": u"Piemēri|Paraugs",
        "given": u"Kad",
        "when": u"Ja",
        "then": u"Tad",
        "and": u"Un",
        "but": u"Bet"
    },
    "nl": {
        "name": u"Dutch",
        "native": u"Nederlands",
        "feature": u"Functionaliteit",
        "background": u"Achtergrond",
        "scenario": u"Scenario",
        "scenario_outline": u"Abstract Scenario",
        "examples": u"Voorbeelden",
        "given": u"Gegeven|Stel",
        "when": u"Als",
        "then": u"Dan",
        "and": u"En",
        "but": u"Maar"
    },
    "no": {
        "name": u"Norwegian",
        "native": u"norsk",
        "feature": u"Egenskap",
        "background": u"Bakgrunn",
        "scenario": u"Scenario",
        "scenario_outline": u"Scenariomal|Abstrakt Scenario",
        "examples": u"Eksempler",
        "given": u"Gitt",
        "when": u"Når",
        "then": u"Så",
        "and": u"Og",
        "but": u"Men"
    },
    "pa": {
        "name": u"Panjabi",
        "native": u"ਪੰਜਾਬੀ",
        "feature": u"ਖਾਸੀਅਤ|ਮੁਹਾਂਦਰਾ|ਨਕਸ਼ ਨੁਹਾਰ",
        "background": u"ਪਿਛੋਕੜ",
        "scenario": u"ਪਟਕਥਾ",
        "scenario_outline": u"ਪਟਕਥਾ ਢਾਂਚਾ|ਪਟਕਥਾ ਰੂਪ ਰੇਖਾ",
        "examples": u"ਉਦਾਹਰਨਾਂ",
        "given": u"ਜੇਕਰ|ਜਿਵੇਂ ਕਿ",
        "when": u"ਜਦੋਂ",
        "then": u"ਤਦ",
        "and": u"ਅਤੇ",
        "but": u"ਪਰ"
    },
    "pl": {
        "name": u"Polish",
        "native": u"polski",
        "feature": u"Właściwość|Funkcja|Aspekt|Potrzeba biznesowa",
        "background": u"Założenia",
        "scenario": u"Scenariusz",
        "scenario_outline": u"Szablon scenariusza",
        "examples": u"Przykłady",
        "given": u"Zakładając(?:, że)?|Mając",
        "when": u"Jeżeli|Jeśli|Gdy|Kiedy",
        "then": u"Wtedy",
        "and": u"Oraz|I",
        "but": u"Ale"
    },
    "pt": {
        "name": u"Portuguese",
        "native": u"português",
        "background": u"Contexto|Cenário de Fundo|Cenario de Fundo|Fundo",
        "feature": u"Funcionalidade|Característica|Caracteristica",
        "scenario": u"Cenário|Cenario",
        "scenario_outline": u"Esquema do Cenário|Esquema do Cenario|Delineação do Cenário|Delineacao do Cenario",
        "examples": u"Exemplos|Cenários|Cenarios",
        "given": u"Dado|Dada|Dados|Dadas",
        "when": u"Quando",
        "then": u"Então|Entao",
        "and": u"E",
        "but": u"Mas"
    },
    "ro": {
        "name": u"Romanian",
        "native": u"română",
        "background": u"Context",
        "feature": u"Functionalitate|Funcționalitate|Funcţionalitate",
        "scenario": u"Scenariu",
        "scenario_outline": u"Structura scenariu|Structură scenariu",
        "examples": u"Exemple",
        "given": u"Date fiind|Dat fiind|Dati fiind|Dați fiind|Daţi fiind",
        "when": u"Cand|Când",
        "then": u"Atunci",
        "and": u"Si|Și|Şi",
        "but": u"Dar"
    },
    "ru": {
        "name": u"Russian",
        "native": u"русский",
        "feature": u"Функция|Функционал|Свойство",
        "background": u"Предыстория|Контекст",
        "scenario": u"Сценарий",
        "scenario_outline": u"Структура сценария",
        "examples": u"Примеры",
        "given": u"Допустим|Дано|Пусть",
        "when": u"Если|Когда",
        "then": u"То|Тогда",
        "and": u"И|К тому же|Также",
        "but": u"Но|А"
    },
    "sv": {
        "name": u"Swedish",
        "native": u"Svenska",
        "feature": u"Egenskap",
        "background": u"Bakgrund",
        "scenario": u"Scenario",
        "scenario_outline": u"Abstrakt Scenario|Scenariomall",
        "examples": u"Exempel",
        "given": u"Givet",
        "when": u"När",
        "then": u"Så",
        "and": u"Och",
        "but": u"Men"
    },
    "sk": {
        "name": u"Slovak",
        "native": u"Slovensky",
        "feature": u"Požiadavka|Funkcia|Vlastnosť",
        "background": u"Pozadie",
        "scenario": u"Scenár",
        "scenario_outline": u"Náčrt Scenáru|Náčrt Scenára|Osnova Scenára",
        "examples": u"Príklady",
        "given": u"Pokiaľ|Za predpokladu",
        "when": u"Keď|Ak",
        "then": u"Tak|Potom",
        "and": u"A|A tiež|A taktiež|A zároveň",
        "but": u"Ale"
    },
    "sl": {
        "name": u"Slovenian",
        "native": u"Slovenski",
        "feature": u"Funkcionalnost|Funkcija|Možnosti|Moznosti|Lastnost|Značilnost",
        "background": u"Kontekst|Osnova|Ozadje",
        "scenario": u"Scenarij|Primer",
        "scenario_outline": u"Struktura scenarija|Skica|Koncept|Oris scenarija|Osnutek",
        "examples": u"Primeri|Scenariji",
        "given": u"Dano|Podano|Zaradi|Privzeto",
        "when": u"Ko|Ce|Če|Kadar",
        "then": u"Nato|Potem|Takrat",
        "and": u"In|Ter",
        "but": u"Toda|Ampak|Vendar"
    },
    "sr-Latn": {
        "name": u"Serbian (Latin)",
        "native": u"Srpski (Latinica)",
        "feature": u"Funkcionalnost|Mogućnost|Mogucnost|Osobina",
        "background": u"Kontekst|Osnova|Pozadina",
        "scenario": u"Scenario|Primer",
        "scenario_outline": u"Struktura scenarija|Skica|Koncept",
        "examples": u"Primeri|Scenariji",
        "given": u"Zadato|Zadate|Zatati",
        "when": u"Kada|Kad",
        "then": u"Onda",
        "and": u"I",
        "but": u"Ali"
    },
    "sr-Cyrl": {
        "name": u"Serbian",
        "native": u"Српски",
        "feature": u"Функционалност|Могућност|Особина",
        "background": u"Контекст|Основа|Позадина",
        "scenario": u"Сценарио|Пример",
        "scenario_outline": u"Структура сценарија|Скица|Концепт",
        "examples": u"Примери|Сценарији",
        "given": u"Задато|Задате|Задати",
        "when": u"Када|Кад",
        "then": u"Онда",
        "and": u"И",
        "but": u"Али"
    },
    "tl": {
        "name": u"Telugu",
        "native": u"తెలుగు",
        "feature": u"గుణము",
        "background": u"నేపథ్యం",
        "scenario": u"సన్నివేశం",
        "scenario_outline": u"కథనం",
        "examples": u"ఉదాహరణలు",
        "given": u"చెప్పబడినది",
        "when": u"ఈ పరిస్థితిలో",
        "then": u"అప్పుడు",
        "and": u"మరియు",
        "but": u"కాని"
    },
    "th": {
        "name": u"Thai",
        "native": u"ไทย",
        "feature": u"โครงหลัก|ความต้องการทางธุรกิจ|ความสามารถ",
        "background": u"แนวคิด",
        "scenario": u"เหตุการณ์",
        "scenario_outline": u"สรุปเหตุการณ์|โครงสร้างของเหตุการณ์",
        "examples": u"ชุดของตัวอย่าง|ชุดของเหตุการณ์",
        "given": u"กำหนดให้",
        "when": u"เมื่อ",
        "then": u"ดังนั้น",
        "and": u"และ",
        "but": u"แต่"
    },
    "tlh": {
        "name": u"Klingon",
        "native": u"tlhIngan",
        "feature": u"Qap|Qu'meH 'ut|perbogh|poQbogh malja'|laH",
        "background": u"mo'",
        "scenario": u"lut",
        "scenario_outline": u"lut chovnatlh",
        "examples": u"ghantoH|lutmey",
        "given": u"ghu' noblu'|DaH ghu' bejlu'",
        "when": u"qaSDI'",
        "then": u"vaj",
        "and": u"'ej|latlh",
        "but": u"'ach|'a"
    },
    "tr": {
        "name": u"Turkish",
        "native": u"Türkçe",
        "feature": u"Özellik",
        "background": u"Geçmiş",
        "scenario": u"Senaryo",
        "scenario_outline": u"Senaryo taslağı",
        "examples": u"Örnekler",
        "given": u"Diyelim ki",
        "when": u"Eğer ki",
        "then": u"O zaman",
        "and": u"Ve",
        "but": u"Fakat|Ama"
    },
    "tt": {
        "name": u"Tatar",
        "native": u"Татарча",
        "feature": u"Мөмкинлек|Үзенчәлеклелек",
        "background": u"Кереш",
        "scenario": u"Сценарий",
        "scenario_outline": u"Сценарийның төзелеше",
        "examples": u"Үрнәкләр|Мисаллар",
        "given": u"Әйтик",
        "when": u"Әгәр",
        "then": u"Нәтиҗәдә",
        "and": u"Һәм|Вә",
        "but": u"Ләкин|Әмма"
    },
    "uk": {
        "name": u"Ukrainian",
        "native": u"Українська",
        "feature": u"Функціонал",
        "background": u"Передумова",
        "scenario": u"Сценарій",
        "scenario_outline": u"Структура сценарію",
        "examples": u"Приклади",
        "given": u"Припустимо|Припустимо, що|Нехай|Дано",
        "when": u"Якщо|Коли",
        "then": u"То|Тоді",
        "and": u"І|А також|Та",
        "but": u"Але"
    },
    "uz": {
        "name": u"Uzbek",
        "native": u"Узбекча",
        "feature": u"Функционал",
        "background": u"Тарих",
        "scenario": u"Сценарий",
        "scenario_outline": u"Сценарий структураси",
        "examples": u"Мисоллар",
        "given": u"Агар",
        "when": u"Агар",
        "then": u"Унда",
        "and": u"Ва",
        "but": u"Лекин|Бирок|Аммо"
    },
    "vi": {
        "name": u"Vietnamese",
        "native": u"Tiếng Việt",
        "feature": u"Tính năng",
        "background": u"Bối cảnh",
        "scenario": u"Tình huống|Kịch bản",
        "scenario_outline": u"Khung tình huống|Khung kịch bản",
        "examples": u"Dữ liệu",
        "given": u"Biết|Cho",
        "when": u"Khi",
        "then": u"Thì",
        "and": u"Và",
        "but": u"Nhưng"
    },
    "zh-CN": {
        "name": u"Chinese simplified",
        "native": u"简体中文",
        "feature": u"功能",
        "background": u"背景",
        "scenario": u"场景|剧本",
        "scenario_outline": u"场景大纲|剧本大纲",
        "examples": u"例子",
        "given": u"假如<|假设<|假定<",
        "when": u"当<",
        "then": u"那么<",
        "and": u"而且<|并且<|同时<",
        "but": u"但是<"
    },
    "zh-TW": {
        "name": u"Chinese traditional",
        "native": u"繁體中文",
        "feature": u"功能",
        "background": u"背景",
        "scenario": u"場景|劇本",
        "scenario_outline": u"場景大綱|劇本大綱",
        "examples": u"例子",
        "given": u"假如<|假設<|假定<",
        "when": u"當<",
        "then": u"那麼<",
        "and": u"而且<|並且<|同時<",
        "but": u"但是<"
    },
    "ur": {
        "name": u"Urdu",
        "native": u"اردو",
        "feature": u"صلاحیت|کاروبار کی ضرورت|خصوصیت",
        "background": u"پس منظر",
        "scenario": u"منظرنامہ",
        "scenario_outline": u"منظر نامے کا خاکہ",
        "examples": u"مثالیں",
        "given": u"اگر|بالفرض|فرض کیا",
        "when": u"جب",
        "then": u"پھر|تب",
        "and": u"اور",
        "but": u"لیکن"
    }
}



from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import time

from .exceptions import MissingStepError


class IVisitor(object):

    __metaclass__ = ABCMeta

    def before_feature(self, node):
        pass

    def after_feature(self, node):
        pass

    def before_scenario(self, node):
        pass

    def after_scenario(self, node):
        pass

    def permute_schedule(self, node):
        return [[0]]

    def step_schedule(self, node):
        return [list(range(len(node.steps)))]

    @abstractmethod
    def visit(self):
        pass  # pragma: nocover


class TestVisitor(IVisitor):
    """ Visits all steps and run step methods. """

    def __init__(self, suite, matcher):
        self._suite = suite
        self._matcher = matcher

    def visit(self, node):
        self._suite.step = node
        start_time = time.time()
        status = ''
        try:
            try:
                node.test_step(self._suite, self._matcher)
            except (MissingStepError, AssertionError):
                status = 'Fail'
                raise
            except Exception:
                status = 'Error'
                raise
            else:
                status = 'OK'
        finally:
            end_time = time.time()
            additional_data = node.additional_data
            additional_data['duration'] = end_time - start_time
            additional_data['status'] = status

    def before_scenario(self, node):
        self._suite.setUp()

    def after_scenario(self, node):
        self._suite.tearDown()

    def permute_schedule(self, node):
        return node.permute_schedule()

    def step_schedule(self, node):
        return node.step_schedule()


class ReportVisitor(IVisitor):
    def __init__(self, suite):
        self._suite = suite
        self._result = ''

    def visit(self, node):
        recon, result = node.to_html()
        if recon[-1] != '\n':
            recon += '\n'  # TODO  clean this outa def reconstruction(s)!
        self._result += recon
        return result

    def after_scenario(self, node):
        if node.result:
            self._result += node.result

    def __str__(self):
        return self._result

    def __unicode__(self):
        return unicode(self._result)


class StepMatcherVisitor(IVisitor):
    """ Visits all steps in order to find missing step methods. """

    def __init__(self, suite, matcher):
        self._suite = suite
        self._not_matched = OrderedDict()
        self._matcher = matcher

    def visit(self, node):
        try:
            node.find_step(self._suite, self._matcher)
        except MissingStepError as e:
            self._not_matched[e.suggest] = True

    def after_feature(self, node):
        suggest = u''.join(self._not_matched.keys())
        if suggest:
            diagnostic = u'Cannot match steps:\n\n%s' % suggest
            self._suite.fail(diagnostic)
#
#
# class TextReportVisitor(IVisitor):
#     def __init__(self, suite, matcher):
#         self._suite = suite
#         self._result = ''
#
#     def visit(self, node):
#         duration = node.additional_data.get('duration', 0)
#         if duration:
#             duration = '#  %.6fs' % duration
#         else:
#             duration = ''
#         self._result += '%-57s %s\n' % (
#             node.reconstruction().strip('\n'),
#             duration,
#         )
#
#     def __str__(self):
#         return self._result
#
#     def __unicode__(self):
#         return unicode(self._result)



class MoreliaError(Exception):
    pass


class MissingStepError(MoreliaError):

    def __init__(self, predicate, suggest, *args, **kwargs):
        self.predicate = predicate
        self.suggest = suggest



# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase
import re
import sys
import os
pwd = os.path.dirname(os.path.realpath(__file__))
morelia_path = os.path.join(pwd, '../morelia')
sys.path.insert(0, morelia_path)
from morelia.base import (Parser, Feature, Scenario, Given, Comment, Step, Row,
                          And, When, Then, TestVisitor, _permute_indices, MissingStepError)
from morelia.matchers import RegexpStepMatcher, MethodNameStepMatcher
from morelia.i18n import TRANSLATIONS
from morelia.utils import to_unicode

#  CONSIDER  same order as morelia.feature, & vice-versa
zones = []
crunks = []
factions = []
elements = []


class MoreliaSuite(TestCase):

    def step_I_have_entered_a_number_into_the_calculator(self, number):
        r'I have entered (\d+) into the calculator'

        if not hasattr(self, 'stack'):
            self.stack = []
        self.stack.append(int(number))

    def step_I_press_add(self):
        self.result = sum(self.stack)

    def step_the_result_should_be_on_the_screen(self, number):
        "the result should be (\d+) on the screen"

        self.assertEqual(int(number), self.result)

    def _get_language(self):
        return None

    def test_feature(self):
        language = self._get_language()
        input = '%s: prevent wild animals from eating us' % self.feature_keyword
        steps = Parser(language=language).parse_feature(input)
        step = steps[0]
        assert step.__class__ == Feature
        self.assertEqual(step.concept, 'Feature')
        self.assertEqual(step.predicate, 'prevent wild animals from eating us')

    def test_scenario(self):
        input = '%s: range free Vegans' % self.scenario_keyword
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        step = steps[0]
        assert step.__class__ == Scenario
        self.assertEqual(step.concept, 'Scenario')
        self.assertEqual(step.predicate, 'range free Vegans')

    def test___scenario(self):
        input = '  %s: with spacies' % self.scenario_keyword
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        step = steps[0]
        assert step.__class__ == Scenario
        self.assertEqual(step.concept, 'Scenario')
        self.assertEqual(step.predicate, 'with spacies')

    def test_given_a_string_with_given_in_it(self):
        input = '%(given)s a string with Given in it   \n%(and)s another string' % {
            'given': self.given_keyword,
            'and': self.and_keyword,
        }
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)  # ^ note the spacies
        step = steps[0]
        assert step.__class__ == Given
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'a string with Given in it')  # <-- note spacies are gone

    def test_given_a_broken_string_with_excess_spacies(self):
        input = '%s a string with spacies and   \n  another string  ' % self.given_keyword
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        step = steps[0]
        assert step.__class__ == Given
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'a string with spacies and\nanother string')

    def test_deal_with_pesky_carriage_returns(self):  # because Morse Code will live forever!
        input = '%s a string with spacies and   \r\n  another string  ' % self.given_keyword
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        step = steps[0]
        assert step.__class__ == Given
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'a string with spacies and\nanother string')

    def test_given_a_string_with_a_line_breaker_followed_by_a_keyword(self):
        input = '%(given)s a string \\\n And another string' % {
            'given': self.given_keyword,
            'and': self.and_keyword,
        }
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        assert 1 == len(steps)
        step = steps[0]
        assert step.__class__ == Given
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'a string \\\n And another string')

    def test_given_a_string_with_a_line_breaker_followed_by_a_keyword_with_stray_spacies(self):
        input = '%(given)s a string \\  \n And another string, without stray spacies' % {
            'given': self.given_keyword,
            'and': self.and_keyword,
        }
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        assert 1 == len(steps)
        step = steps[0]
        assert step.__class__ == Given
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'a string \\\n And another string, without stray spacies')

    def test_feature_with_scenario(self):
        input = '''%(feature)s: Civi-lie-zation
                   %(scenario)s: starz upon tharz bucks''' % {
                'feature': self.feature_keyword,
                'scenario': self.scenario_keyword,
        }
        language = self._get_language()
        steps = Parser(language=language).parse_feature(input)
        step = steps[0]
        assert step.__class__ == Feature
        self.assertEqual(step.concept, 'Feature')
        self.assertEqual(step.predicate, 'Civi-lie-zation')
        step = steps[1]
        assert step.__class__ == Scenario
        self.assertEqual(step.concept, 'Scenario')
        self.assertEqual(step.predicate, 'starz upon tharz bucks')

    # def test_feature_with_lone_comment(self):
    #     input = '''Feature: The Sacred White Llama of the Inca\r
    #                #  I are a comment'''
    #     steps = Parser().parse_feature(input)
    #     assert steps[0].__class__ == Feature

    #     step = steps[1]
    #     assert step.__class__ == Comment
    #     self.assertEqual(step.concept, 'Comment')
    #     self.assertEqual(step.predicate, 'I are a comment')

    def test_feature_with_lone_comment(self):
        input = 'i be a newbie feature'
        language = self._get_language()
        p = Parser(language=language)

        try:
            p.parse_feature(input)
            assert False  # should fail!
        except SyntaxError as e:
            e = e.args[0]
            self.assert_regex_contains(r'File "None", line 1, in', e)
            self.assert_regex_contains(r'\?\?\?: i be a newbie feature', e)
            try:
                feature_name = TRANSLATIONS[language].get('feature', u'Feature')
            except KeyError:
                feature_name = 'Feature'
            else:
                feature_name = feature_name.replace('|', ' or ')
            self.assert_regex_contains(r'feature files must start with a %s' % feature_name, e)

    def test_feature_with_long_comment(self):  # ERGO how to detect shadowed test cases??
        language = self._get_language()
        p = Parser(language=language)

        input = '''%s: The Sacred Giant Mosquito of the Andes
                   #  at http://www.onagocag.com/nazbird.jpg
                        so pay no attention to the skeptics!''' % self.feature_keyword
        try:
            p.parse_feature(input)
            assert False  # should raise a SyntaxError
        except SyntaxError as e:
            self.assert_regex_contains('linefeed in comment', str(e))
            self.assert_regex_contains('line 2', str(e))

        steps = p.steps
        assert steps[0].__class__ == Feature
        step = steps[1]
        assert step.__class__ == Comment
        self.assertEqual(step.concept, 'Comment')

    def pet_scenario(self):
        return '''%(scenario)s: See all vendors
                      %(given)s I am logged in as a user in the administrator role
                        %(and)s There are 3 vendors
                       %(when)s I go to the manage vendors page
                       %(then)s I should see the first 3 vendor names''' % {
            'scenario': self.scenario_keyword,
            'given': self.given_keyword,
            'and': self.and_keyword,
            'when': self.when_keyword,
            'then': self.then_keyword,
        }

    def test_parse_scenario(self):
        scenario = self.pet_scenario()
        language = self._get_language()
        steps = Parser(language=language).parse_feature(scenario)
        step_0, step_1, step_2, step_3, step_4 = steps
        self.assertEqual(step_0.concept, 'Scenario')
        self.assertEqual(step_0.predicate, 'See all vendors')
        self.assertEqual(step_1.concept, 'Given')
        self.assertEqual(step_1.predicate, 'I am logged in as a user in the administrator role')
        self.assertEqual(step_2.concept, 'And')
        self.assertEqual(step_2.predicate, 'There are 3 vendors')
        self.assertEqual(step_3.concept, 'When')
        self.assertEqual(step_3.predicate, 'I go to the manage vendors page')
        self.assertEqual(step_4.concept, 'Then')
        self.assertEqual(step_4.predicate, 'I should see the first 3 vendor names')

    def test_strip_predicates(self):
        language = self._get_language()
        step = Parser(language=language).parse_feature('  %s   gangsta girl   \t     ' % self.given_keyword)[0]
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'gangsta girl')

    def test_bond_predicates(self):
        return  # CONSIDER  why test_strip_predicates passes and this croaks???
        language = self._get_language()
        step = Parser(language=language).parse_feature('  %s\n   elf quest   \t     ' % self.given_keyword)[0]
        self.assertEqual(step.concept, 'Given')
        self.assertEqual(step.predicate, 'elf quest')

    def test_scenarios_link_to_their_steps(self):
        language = self._get_language()
        steps = Parser(language=language).parse_feature(self.pet_scenario())
        scenario, step_1, step_2, step_3, step_4 = steps
        self.assertEqual([step_1, step_2, step_3, step_4], scenario.steps)

    def test_how_to_identify_trees_from_quite_a_long_distance_away(self):
        assert Given != Step
        assert issubclass(Given, Step)
        assert issubclass(Given, Given)
        assert not issubclass(Scenario, Given)

    def test_i_look_like(self):
        self.assertEqual('Step', Step().i_look_like())
        self.assertEqual('Given', Given().i_look_like())
        self.assertEqual('\\|', Row().i_look_like())

    def test_evaluate_step_by_name(self):
        step = Given()._parse('my milkshake')
        self.youth = 'girls'
        matcher = self._get_default_machers()
        step.evaluate(self, matcher)
        self.assertEqual('boys', self.youth)

# ####  row zone  #################################

    def test_Row_parse(self):
        sauce = 'buddha | brot |'
        row = Row()
        row._parse(sauce, [])
        assert row.predicate == sauce

    def test_parse_feature_Row(self):
        language = self._get_language()
        p = Parser(language=language)
        p.parse_features(''' | piggy | op |''')
        assert Row == p.steps[0].__class__
        assert p.steps[0].predicate == 'piggy | op |'

    def test_Scenes_count_Row_dimensions(self):
        self.assemble_scene_table()
        dims = self.table_scene.steps[0].steps[0].count_Row_dimensions()
        self.assertEqual([2, 3], dims)

    def test_Scenes_count_more_Row_dimensions(self):
        self.assemble_scene_table('Step whatever\n')
        dims = self.table_scene.steps[0].steps[0].count_Row_dimensions()
        self.assertEqual([2, 0, 3], dims)

    def test_permutate(self):
        expect = [(0, 0, 0), (0, 0, 1), (0, 0, 2), (0, 1, 0), (0, 1, 1),
                  (0, 1, 2), (0, 2, 0), (0, 2, 1), (0, 2, 2), (0, 3, 0),
                  (0, 3, 1), (0, 3, 2)]
        self.assertEqual(expect, _permute_indices([0, 4, 3]))
        expect = [(0, 0, 0)]
        self.assertEqual(expect, _permute_indices([1, 1, 1]))
        expect = [(0, 0, 0), (0, 0, 1)]
        self.assertEqual(expect, _permute_indices([1, 1, 2]))

    def assemble_scene_table_source(self, moar=''):
        return '''%(feature)s: permute tables
                       %(scenario)s: turn one feature into many
                           %(given)s party <zone>
                                | zone  |
                                | beach |
                                | hotel |
                           %(moar)s%(then)s hearty <crunk>
                                | crunk |
                                | work  |
                                | mall  |
                                | jail  |''' % {
            'moar': moar,
            'feature': self.feature_keyword,
            'scenario': self.scenario_keyword,
            'given': self.given_keyword,
            'then': self.then_keyword,
        }

    def assemble_scene_table(self, moar=''):
        scene = self.assemble_scene_table_source(moar)
        language = self._get_language()
        p = Parser(language=language)
        self.table_scene = p.parse_features(scene)

    def test_permute_schedule(self):
        expect = _permute_indices([2, 0, 3])  # NOTE:  by rights, 0 should be -1
        self.assemble_scene_table('Step you betcha\n')
        scenario = self.table_scene.steps[0].steps[0]
        schedule = scenario.permute_schedule()
        self.assertEqual(expect, schedule)

    def test_evaluate_permuted_schedule(self):
        self.assemble_scene_table('Step flesh is weak\n')
        scenario = self.table_scene.steps[0].steps[0]
        steps_num = len(scenario.steps)
        matcher = RegexpStepMatcher(self).add_matcher(MethodNameStepMatcher(self))
        visitor = TestVisitor(self, matcher)
        global crunks, zones
        crunks = []
        zones = []
        scenario.row_indices = [1, 0, 2]
        scenario.evaluate_test_case(visitor, range(steps_num))
        self.assertEqual('hotel', visitor._suite.got_party_zone)
        self.assertEqual('jail', visitor._suite.got_crunk)

    def test_another_two_dimensional_table(self):
        global crunks, zones
        crunks = []
        zones = []
        scene = self.assemble_scene_table_source('Step my milkshake brings all the boys to the yard\n')
        language = self._get_language()
        Parser(language=language).parse_features(scene).evaluate(self)
        self.assertEqual(['work', 'mall', 'jail', 'work', 'mall', 'jail'], crunks)
        self.assertEqual(['beach', 'beach', 'beach', 'hotel', 'hotel', 'hotel'], zones)

    def assemble_multiple_whens(self):
        return '''%(scenario)s: Split When Blocks
                      %(given)s some setup
                       %(when)s a first trigger occurs
                       %(then)s something good happens
                       %(when)s another trigger occurs
                       %(then)s something else happens''' % {
            'scenario': self.scenario_keyword,
            'given': self.given_keyword,
            'when': self.when_keyword,
            'then': self.then_keyword,
        }

#  TODO  what happens when a scene has 2 tables and it pulls keywords from either one?
#  TODO  what happens when a scene has different tables in different When blocks?

    def test_parse_multiple_whens(self):
        scenario = self.assemble_multiple_whens()
        language = self._get_language()
        steps = Parser(language=language).parse_feature(scenario)  # TODO  test that a non-double-When Scenario gives a flat schedule
        scene, step_1, step_2, step_3, step_4, step_5 = steps
        assert scene.concept == 'Scenario'
        assert [[0, 1, 2], [0, 3, 4]] == scene.step_schedule()  # TODO  mix step schedules and row schedules! (-:
        # TODO  better object model! assert 2 == step_0.count_whens()

    def test_harvest(self):
        r = Row()

        def harvest(predicate):
            r.predicate = predicate
            return r.harvest()

        self.assertEqual(['crock', 'of'], harvest('crock | of'))
        self.assertEqual(['crock', 'of'], harvest('crock | of |'))

        # CONSIDER  document you gotta take the \ out yourself

        self.assertEqual(['crane \| wife', 'three'], harvest('crane \| wife | three'))

    def step_party_zone(self, zone):  # CONSIDER  prevent collision with another "step_party"
        r'party (\w+)'  # CONSIDER  illustrate how the patterns here form testage too

        self.got_party_zone = zone
        global zones
        zones.append(zone)

    def step_flesh_is_weak(self):
        pass

    def step_hearty_crunk_(self, crunk):
        r'hearty (\w+)'

        global crunks
        crunks.append(crunk)
        self.got_crunk = crunk

#  TODO  squeak if the table item ain't found
#  TODO  respect the tests' verbosity levels
#  CONSIDER  note that default arguments on steps are permitted!
#  CONSIDER  parse the || as Json/Yaml? - permit gaps & comments in tables

    def test_Rows_find_step_parents(self):
        self.assemble_scene_table()
        given, then, = self.table_scene.steps[0].steps[0].steps
        self.assertEqual(Row, given.steps[0].__class__)
        self.assertEqual(Row, then.steps[0].__class__)
        self.assertEqual('zone  |', given.steps[0].predicate)
        self.assertEqual('crunk |', then.steps[0].predicate)

    def assemble_short_scene_table(self):
        #  TODO  warn about bad but permitted style -- with | columns | out of order!
            #  TODO  reporter should beautify || markers!
        return '''%(feature)s: the smoker you drink
                    %(scenario)s: the programmer you get
                      %(given)s party <element> from <faction>

                                | faction   | element  |

                                | Pangolin  | Pangea   |
                                | Glyptodon | Laurasia |''' % {
            'feature': self.feature_keyword,
            'scenario': self.scenario_keyword,
            'given': self.given_keyword,
        }

    def test_two_dimensional_table(self):
        global elements, factions
        elements = []
        factions = []
        language = self._get_language()
        Parser(language=language).parse_features(self.assemble_short_scene_table()).evaluate(self)
        self.assertEqual([['Pangolin', 'Glyptodon'], ['Pangea', 'Laurasia']], [factions, elements])

    def test_two_dimensional_table_reconstruction(self):
        language = self._get_language()
        p = Parser(language=language).parse_features(self.assemble_short_scene_table())
        step = p.steps[0].steps[0].steps[0]
        self.assertEqual(step.concept + ': ' + step.predicate, step.reconstruction().strip())

    def step_party_element_from_faction(self, element, faction):
        r'party (\w+) from (\w+)'
        # TODO  don't default to this "party <element> in <faction>"

        global elements, factions
        factions.append(faction)
        elements.append(element)

    def step_my_milkshake(self, youth='boys', article='the'):
        r'my milkshake brings all the (boys|girls) to (.*) yard'

        self.youth = youth

    def step_exceptional(self):
        x = 1 / 0  # noqa guilty pleasure for programmers!

    def test_handle_exceptions(self):
        s = Step()

        s.concept = 'Given'
        s.predicate = 'exceptional'
        s.line_number = 42
        matcher = RegexpStepMatcher(self).add_matcher(MethodNameStepMatcher(self))
        visitor = TestVisitor(self, matcher)
        matcher = self._get_default_machers()

        try:
            s.test_step(visitor, matcher)
            assert False  # should raise!
        except ZeroDivisionError as e:
            assert 'Given: exceptional' in str(e)

    def test_find_step_by_name(self):
        step = Given()._parse('my milkshake')
        matcher = self._get_default_machers()
        method, args, kwargs = step.find_step(self, matcher)
        expect = self.step_my_milkshake
        self.assertEqual(expect, method)

    def test_find_step_by_doc_string(self):
        step = And()._parse('my milkshake brings all the boys to the yard')
        matcher = self._get_default_machers()
        method, args, kwargs = step.find_step(self, matcher)
        expect = self.step_my_milkshake
        self.assertEqual(expect, method)

    def test_find_step_with_match(self):
        step = When()._parse('my milkshake brings all the girls to the yard')
        matcher = self._get_default_machers()
        method, args, kwargs = step.find_step(self, matcher)
        self.assertEqual(('girls', 'the'), args)

    def test_step_not_found(self):
        step = Then()._parse('not there')
        matcher = self._get_default_machers()
        self.assertRaises(MissingStepError, step.find_step, self, matcher)

    def step_fail_without_enough_function_name(self):
        step = And()._parse('my milk')
        matcher = self._get_default_machers()
        self.assertRaises(MissingStepError, step.find_step, self, matcher)

    def step_fail_step_without_enough_doc_string(self):
        step = Given()._parse("brings all the boys to the yard it's better than yours")
        matcher = self._get_default_machers()
        self.assertRaises(MissingStepError, step.find_step, self, matcher)

    def step_evaluate_step_by_doc_string(self):
        step = Given()._parse('my milkshake brings all the girls to a yard')
        self.youth = 'boys'
        matcher = self._get_default_machers()
        step.evaluate(self, matcher)
        self.assertEqual('girls', self.youth)  # Uh...

    def step_multiline_predicate(self):
        feature = '%s umma\ngumma' % self.given_keyword
        language = self._get_language()
        steps = Parser(language=language).parse_feature(feature)
        self.assertEqual('umma\ngumma', steps[0].predicate)

    def test_step_multiline_predicate(self):
        feature = '%s multiline predicate' % self.when_keyword
        language = self._get_language()
        steps = Parser(language=language).parse_feature(feature)
        matcher = self._get_default_machers()
        steps[0].evaluate(self, matcher)

# CONSIDER use the suite._testMethodDoc to get the doc()! (and what can it do??)
# CONSIDER  use suite.fail instead of raise

#    def test_evaluate_unfound(self):  CONSIDER   real test outa this
#       Parser().parse_file(pwd + '/nada.feature').evaluate(self)

    def test_record_filename(self):
        language = self._get_language()
        filename = pwd + '/morelia%s.feature' % (language or '')
        thang = Parser(language=language).parse_file(filename)
        feature = thang.steps[0]
        assert feature.__class__ == Feature
        assert feature.filename == filename
        step = feature.steps[3].steps[1]
        assert filename == step.get_filename()

    def test_format_faults_like_python_errors(self):
        language = self._get_language()
        filename = pwd + '/morelia%s.feature' % (language or '')
        thang = Parser(language=language).parse_file(filename)
        step = thang.steps[0].steps[3].steps[1]
        assert filename == step.get_filename()
        omen = 'The Alpine glaciers move'
        diagnostic = step.format_fault(omen)
        parent_reconstruction = step.parent.reconstruction().replace('\n', '\\n')
        reconstruction = step.reconstruction().replace('\n', '\\n')

        expect = '\n  File "%s", line %s, in %s\n    %s\n%s' % \
            (step.get_filename(), step.line_number, parent_reconstruction, reconstruction, omen)

        assert expect == diagnostic

    def test_evaluate_file(self):
        language = self._get_language()
        thang = Parser(language=language).parse_file(pwd + '/morelia%s.feature' % (language or ''))
        thang.evaluate(self)

    def setUp(self):
        self.culture = []
        self.feature_keyword = 'Feature'
        self.scenario_keyword = 'Scenario'
        self.given_keyword = 'Given'
        self.then_keyword = 'Then'
        self.when_keyword = 'When'
        self.and_keyword = 'And'

    def step_adventure_of_love_love_and_culture_(self, culture):
        r'adventure of love - love and (.+)'

        self.culture.append(culture)
        self.concept = self.step.concept

    def step_Moralia_evaluates_this(self):
        pass

    def step_culture_contains(self, arguments):
        r'"culture" contains (.*)'

        self.assertEqual(1, arguments.count(self.culture[0]))
        self.assertEqual(1, len(self.culture))

    def _xml_to_tree(self, xml):
        from lxml import etree
        self._xml = xml

        try:
            if '<html' in xml[:200]:
                return etree.HTML(xml)
            else:
                return etree.XML(xml)

        except ValueError:  # TODO don't rely on exceptions for normal control flow
            tree = xml
            self._xml = str(tree)  # CONSIDER does this reconstitute the nested XML ?
            return tree

    def assert_xml(self, xml, xpath, **kw):
        'Check that a given extent of XML or HTML contains a given XPath, and return its first node'

        tree = self._xml_to_tree(xml)
        nodes = tree.xpath(xpath)
        self.assertTrue(len(nodes) > 0, xpath + ' not found in ' + self._xml)
        node = nodes[0]
        if kw.get('verbose', False):
            self.reveal_xml(node)  # "here have ye been? What have ye seen?"--Morgoth
        return node

    def test_report_file(self):
        language = self._get_language()
        thang = Parser(language=language).parse_file(pwd + '/morelia%s.feature' % (language or ''))
        div_count = len(thang.steps[0].steps)  # CONSIDER  this off-by-one and on-by-one; dunno why, needs fixed

        rep = thang.report(self)._result
        once = 'when did Bow Wow Wow become classic rock'
        assert 1 == rep.count(once)

        html = '<xml>' + rep + '</xml>'
        # open('/tmp/yo.html', 'w').write(html.encode('utf-8'))
        # ERGO assert_xml with <html> forgives - crack down on that!

        self.assert_xml(html, '/xml[ count(descendant::div) > %i ]' % (div_count - 1))
        #  os.system('firefox /home/phlip/morelia/yo.html &')
        # os.system('konqueror  /home/phlip/morelia/yo.html &')

    def step_a_feature_file_with_contents(self, file_contents):
        r'a feature file with "([^"]+)"'
        self.file_contents = file_contents

    # def step_wikked_(self):
    # '''wikked!
    # '''
    #  TODO get rid of linefeeds in the suggestions
    # assert False

    def step_Moralia_evaluates_the_file(self):
        self.diagnostic = None
        self.steps = []

        try:
            language = self._get_language()
            p = Parser(language=language)
            self.file_contents.replace('\\#', '#')  # note - this is how to unescape characters - DIY
            p.parse_features(self.file_contents).evaluate(self)
            self.steps = p.steps
        except AssertionError as e:
            self.diagnostic = str(e)

    def step_it_prints_a_diagnostic(self, sample):
        r'it prints a diagnostic containing "([^"]+)"'

        self.assert_regex_contains(re.escape(sample), self.diagnostic)

    def step_the_second_line_contains(self, docstring):
        r'the second line contains "([^"]+)"'

        self.assert_regex_contains(re.escape(docstring), self.diagnostic)

    def step_it_contains_1_step(self):
        r'it contains 1 step'

        self.assertEqual(1, len(self.steps))

    def step_the_step_concept_is_(self, concept):
        r'the step concept is (.+)'

        self.assertEqual(concept, self.concept)

    def step_a_source_file_with_a_Given_(self, predicate):
        r'a source file with a (.+)'

        self.predicate = predicate.replace('\\n', '\n')

    def step_we_evaluate_the_file(self):
        r'we evaluate the file'

        matcher = self._get_default_machers()
        self.suggestion, self.extra_arguments = matcher._suggest_doc_string(self.predicate)

    def step_we_convert_it_into_a_(self, suggestion):
        r'we convert it into a (.+)'

        self.assertEqual(suggestion, self.suggestion)

    def step_add_extra_arguments(self, extra=''):  # TODO  blank columns should exist!
        r'add (.+) arguments'

        self.assertEqual(extra, self.extra_arguments)

    def step_a_file_contains_statements_produce_diagnostics_(self, statements, diagnostics):
        r'a file contains (.+), it produces (.+)'

        try:
            statements = statements.replace('\\n', '\n')  # CONSIDER  document this is how you paint linefeedage
            statements = statements.replace('\\', '')  # CONSIDER document this is how you paint reserved words
            # diagnostics = diagnostics.replace('\\', '')  #  CONSIDER  document this is how you escape pipes
            # print len(self.step.steps)  #  CONSIDER  document this as the way to hit the whole table
            language = self._get_language()
            p = Parser(language=language).parse_features(statements)
            p.evaluate(self)
            raise Exception('we expect syntax errors here')
        except (SyntaxError, AssertionError) as e:
            e = e.args[0]
            beef, squeak = diagnostics.split(', line ')
            squeak = 'line ' + squeak
            self.assert_regex_contains(re.escape(beef), e)
            self.assert_regex_contains(re.escape(squeak), e)

    def step_errors(self):
        raise SyntaxError('no, you!')

    def assert_regex_contains(self, pattern, string, flags=None):
        flags = flags or 0
        pattern = to_unicode(pattern)
        string = to_unicode(string)
        diagnostic = u'"%s" not found in "%s"' % (pattern, string)
        self.assertTrue(re.search(pattern, string, flags) is not None, diagnostic)

    def _get_default_machers(self):
        docstring_matcher = RegexpStepMatcher(self)
        method_name_matcher = MethodNameStepMatcher(self)
        docstring_matcher.add_matcher(method_name_matcher)
        return docstring_matcher


# Scenario: Leading # marks comment lines.
    # (Warning: Only leading marks are respected for now!)
    # Given a feature file with "When something
    # \# Given nothing"
    # When Moralia evaluates the file
    # Then it contains 1 step
    # And the first step contains "something"
    # And the first step does not contain "nothing"

# CONSIDER  count test cases correctly regarding entire batch - if pyUnit's architecture permits


class PLMoreliaSuite(MoreliaSuite):

    def setUp(self):
        self.culture = []
        self.feature_keyword = 'Właściwość'
        self.scenario_keyword = 'Scenariusz'
        self.given_keyword = 'Zakładając, że'
        self.then_keyword = 'Wtedy'
        self.when_keyword = 'Gdy'
        self.and_keyword = 'I'

    def _get_language(self):
        return 'pl'

    def test_language_directive(self):
        input = '# language: pl\n%s: prevent wild animals from eating us' % self.feature_keyword
        steps = Parser().parse_feature(input)
        step = steps[0]
        assert step.__class__ == Feature
        self.assertEqual(step.concept, 'Feature')
        self.assertEqual(step.predicate, 'prevent wild animals from eating us')


if __name__ == '__main__':
    unittest.main()  # NOTE  this seems to return the correct shell value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase

from mock import sentinel, Mock

from morelia.base import Parser, DEFAULT_LANGUAGE, Morelia, MissingStepError, Step


class ParserParseLanguageDirectiveTestCase(TestCase):
    """ Test :py:meth:`Parser.parse_language_directive`. """

    def test_should_return_true_on_language_directive_line(self):
        """ Scenario: line with language directive """
        # Arrange
        obj = Parser()
        for lang in ['en', 'pl', 'ja']:
            line = '# language: %s' % lang
            # Act
            result = obj.parse_language_directive(line)
            # Assert
            self.assertTrue(result)
            self.assertEqual(obj.language, lang)

    def test_should_return_false_if_no_language_directive(self):
        """ Scenario: line without language directive """
        # Arrange
        obj = Parser()
        lines = [
            '# language: ',  # missing language
            'Feature:',  # not a language directive
            'language: pl',  # missing comment
            '# comment',  # comment
        ]
        for line in lines:
            # Act
            result = obj.parse_language_directive(line)
            # Assert
            self.assertFalse(result)
            self.assertEqual(obj.language, DEFAULT_LANGUAGE)


class MoreliaReconstructionTestCase(TestCase):
    """ Test :py:meth:`Morelia.reconstruction`. """

    def test_should_reconstruct_unicode(self):
        """ Scenario: unicode input """
        # Arrange
        obj = Morelia()
        obj.concept = '???'
        obj.predicate = u'zażółć gęślą jaźń'
        # Act
        result = obj.reconstruction()
        # Assert
        self.assertEqual(result, u'???: zażółć gęślą jaźń\n')

    def test_should_reconstruct_utf8(self):
        """ Scenario: utf8 input """
        # Arrange
        obj = Morelia()
        obj.concept = '???'
        obj.predicate = 'zażółć gęślą jaźń'
        # Act
        result = obj.reconstruction()
        # Assert
        self.assertEqual(result, u'???: zażółć gęślą jaźń\n')


class MissingStepErrorTestCase(TestCase):
    """ Test :py:meth:`MissingStepError`. """

    def test_should_create_exception(self):
        """ Scenario: exception with step and suggest """
        # Arrange
        # Act
        result = MissingStepError('predicate line', 'def step_predicate_line(self):\n    pass')
        # Assert
        self.assertTrue(result is not None)
        self.assertEqual(result.predicate, 'predicate line')


class StepFindStepTestCase(TestCase):
    """ Test :py:meth:`Step.find_step`. """

    def test_should_find_method(self):
        """ Scenario: method found """
        # Arrange
        obj = Step()
        obj.predicate = 'method'
        # Act
        matcher = Mock()
        matcher.find.return_value = (sentinel.method, [], {})
        result, args, kwargs = obj.find_step(sentinel.suite, matcher)
        # Assert
        self.assertEqual(result, sentinel.method)

    def test_should_not_find_method(self):
        """ Scenario: not found """
        # Arrange
        obj = Step()
        obj.predicate = u'some_method'
        matcher = Mock()
        matcher.find.return_value = (None, [], {})
        self.assertRaises(MissingStepError, obj.find_step, sentinel.suite, matcher)



from setuptools import setup, find_packages

with open('README.rst') as file:
    long_description = file.read()

setup(
    name='Morelia',
    version="0.3.0",
    description='for "Behavior Driven Development" (BDD) -- a client-facing scripting language to put the squeeze on all your features',
    long_description=long_description,
    url='http://c2.com/cgi/wiki?MoreliaViridis',
    author='Phlip',
    author_email='phlip2005@gmail.com',
    license="MIT",
    keywords="test bdd behavior",
    packages=find_packages('src', exclude=['example*']),
    package_dir={'': 'src'},
    install_requires=[
        'six',
        'parse',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Testing',
    ],
    test_suite="morelia.tests",
)



from abc import ABCMeta, abstractmethod
import re
import unicodedata

import parse

from .utils import to_unicode


class IStepMatcher(object):
    """ Matches methods to steps.

    Chain of Responsibility.
    """

    __metaclass__ = ABCMeta

    def __init__(self, suite, step_pattern='^step_'):
        self._suite = suite
        self._matcher = re.compile(step_pattern)
        self._next = None

    def _get_all_step_methods(self):
        match = self._matcher.match
        return [method_name for method_name in dir(self._suite) if match(method_name)]

    def add_matcher(self, matcher):
        """ Add new matcher at end of CoR

        :param matcher: matcher to add
        :returns: self
        """
        if self._next is None:
            self._next = matcher
        else:
            self._next.add_matcher(matcher)
        return self

    def find(self, predicate, augmented_predicate, step_methods=None):
        if step_methods is None:
            step_methods = self._get_all_step_methods()
        method, args, kwargs = self.match(predicate, augmented_predicate, step_methods)
        if method:
            return method, args, kwargs
        if self._next is not None:
            return self._next.find(predicate, augmented_predicate, step_methods)
        return None, (), {}

    @abstractmethod
    def match(self, predicate, augmented_predicate, step_methods):
        pass  # pragma: nocover

    @abstractmethod
    def suggest(self, predicate):
        pass  # pragma: nocover

    def _suggest_doc_string(self, predicate):
        extra_arguments = ''
        predicate = predicate.replace("'", r"\'").replace('\n', r'\n')

        # support for tables
        extra_arguments += self._add_extra_args(r'\<(.+?)\>', predicate)
        predicate = self.replace_table_placeholder(predicate)

        # support for variables
        extra_arguments += self._add_extra_args(r'"(.+?)"', predicate)
        predicate = self.replace_variable_placeholder(predicate)

        predicate = re.sub(r' \s+', r'\s+', predicate)
        return "ur'%s'" % predicate, extra_arguments

    def replace_variable_placeholder(self, predicate):
        predicate = re.sub(r'".+?"', '"([^"]+)"', predicate)
        return predicate

    def replace_table_placeholder(self, predicate):
        predicate = re.sub(r'\<.+?\>', '(.+)', predicate)
        return predicate

    def _add_extra_args(self, matcher, predicate):
        args = re.findall(matcher, predicate)
        return ''.join(', ' + self.slugify(arg) for arg in args)

    def slugify(self, predicate):
        predicate = to_unicode(predicate)
        predicate = unicodedata.normalize('NFD', predicate).encode('ascii', 'replace').decode('utf-8')
        predicate = predicate.replace(u'??', u'_').replace(u'?', u'')
        return re.sub(u'[^\w]+', u'_', predicate, re.U).strip('_')


class MethodNameStepMatcher(IStepMatcher):

    def match(self, predicate, augmented_predicate, step_methods):
        clean = re.sub(r'[^\w]', '_?', predicate)
        pattern = '^step_' + clean + '$'
        regexp = re.compile(pattern)
        step_methods = [method for method in step_methods if regexp.match(method)]
        for method_name in step_methods:
            method = self._suite.__getattribute__(method_name)
            return method, (), {}
        return None, (), {}

    def suggest(self, predicate):
        """ Suggest method definition.

        :param str predicate: step predicate
        :returns: suggested method definition
        """
        doc_string, extra_arguments = self._suggest_doc_string(predicate)
        method_name = self.slugify(predicate)
        indent = ' ' * 4
        suggest = u'%(indent)sdef step_%(method_name)s(self%(args)s):\n\n        # code\n        pass\n\n' % {
            'indent': indent,
            'method_name': method_name,
            'args': extra_arguments,
        }
        return suggest


class RegexpStepMatcher(IStepMatcher):

    def match(self, predicate, augmented_predicate, step_methods):
        for method_name in step_methods:
            method = self._suite.__getattribute__(method_name)
            doc = method.__doc__
            if not doc:
                continue
            doc = re.compile('^' + doc + '$')  # CONSIDER deal with users who put in the ^$
            m = doc.match(augmented_predicate)

            if m:
                kwargs = m.groupdict()
                if not kwargs:
                    args = m.groups()
                else:
                    args = ()
                return method, args, kwargs
        return None, (), {}

    def suggest(self, predicate):
        """ Suggest method definition.

        :param str predicate: step predicate
        :returns: suggested method definition
        """
        doc_string, extra_arguments = self._suggest_doc_string(predicate)
        method_name = self.slugify(predicate)
        indent = ' ' * 4
        suggest = u'%(indent)sdef step_%(method_name)s(self%(args)s):\n        %(doc_string)s\n\n        # code\n        pass\n\n' % {
            'indent': indent,
            'method_name': method_name,
            'args': extra_arguments,
            'doc_string': doc_string
        }
        return suggest


class ParseStepMatcher(IStepMatcher):

    def match(self, predicate, augmented_predicate, step_methods):
        for method_name in step_methods:
            method = self._suite.__getattribute__(method_name)
            doc = method.__doc__
            if not doc:
                continue
            match = parse.parse(doc, augmented_predicate)
            if match:
                args = match.fixed
                kwargs = match.named
                return method, tuple(args), kwargs
        return None, (), {}

    def suggest(self, predicate):
        """ Suggest method definition.

        :param str predicate: step predicate
        :returns: suggested method definition
        """
        doc_string, extra_arguments = self._suggest_doc_string(predicate)
        method_name = self.slugify(predicate)
        indent = ' ' * 4
        suggest = u'%(indent)sdef step_%(method_name)s(self%(args)s):\n        %(doc_string)s\n\n        # code\n        pass\n\n' % {
            'indent': indent,
            'method_name': method_name,
            'args': extra_arguments,
            'doc_string': doc_string
        }
        return suggest

    def replace_variable_placeholder(self, predicate):
        predicate = re.sub(r'"(.+?)"', r'"{\1}"', predicate)
        return predicate

    def replace_table_placeholder(self, predicate):
        predicate = re.sub(r'\<(.+?)\>', r'{\1}', predicate)
        return predicate



def to_unicode(text):
    ''' Try convert to unicode independently on python version. '''
    try:
        text = text.decode('utf-8')
    except (UnicodeDecodeError, UnicodeEncodeError, AttributeError):
        pass
    return text



# -*- coding: utf-8 -*-
#            __  __                _ _
#           |  \/  | ___  _ __ ___| (_) __ _
#           | |\/| |/ _ \| '__/ _ \ | |/ _` |
#           | |  | | (_) | | |  __/ | | (_| |
#           |_|  |_|\___/|_|  \___|_|_|\__,_|
#                             o        o     |  o
#                                 ,_       __|      ,
#                        |  |_|  /  |  |  /  |  |  / \_
#                         \/  |_/   |_/|_/\_/|_/|_/ \/
import itertools
import re
from abc import ABCMeta

from six import moves

from .exceptions import MissingStepError
from .i18n import TRANSLATIONS
from .matchers import MethodNameStepMatcher, RegexpStepMatcher, ParseStepMatcher
from .utils import to_unicode
from .visitors import TestVisitor, ReportVisitor, StepMatcherVisitor


__version__ = '0.3.0'


#  TODO  what happens with blank table items?
#  ERGO  river is to riparian as pond is to ___?

LANGUAGE_RE = re.compile(r'^# language: (\w+)')
DEFAULT_LANGUAGE = 'en'


class INode(object):

    __metaclass__ = ABCMeta

    def test_step(self, suite, matcher):
        pass

    def find_step(self, suite, matcher):
        pass


class Morelia(INode):

    def __init__(self, language=DEFAULT_LANGUAGE):
        self.parent = None
        self.language = language
        self.additional_data = {}

    def _parse(self, predicate, list=[], line_number=0):
        self.concept = str(self)
        self.predicate = predicate
        self.steps = []
        self.line_number = line_number

#  TODO  escape the sample regices already!
#  and the default code should be 'print <arg_names, ... >'

        for s in list[::-1]:
            mpt = self.my_parent_type()

            try:
                if issubclass(s.__class__, mpt):
                    s.steps.append(self)  # TODO  squeek if can't find parent
                    self.parent = s
                    break
            except TypeError:
                self.enforce(False, 'Only one Feature per file')  # CONSIDER  prevent it don't trap it!!!

        return self

    def prefix(self):
        return ''

    def my_parent_type(self):
        return None

    def _my_regex(self):
        # TODO  calculate name inside
        name = self.i_look_like()
        return r'\s*(' + name + '):?\s+(.*)'

    def evaluate_steps(self, visitor):
        visitor.visit(self)
        for step in self.steps:
            step.evaluate_steps(visitor)

    def i_look_like(self):
        class_name = str(self)
        name = class_name.lower()
        return TRANSLATIONS[self.language].get(name, class_name)

    def count_dimensions(self):
        ''' Get number of rows. '''
        return sum([step.count_dimension() for step in self.steps])

    def count_dimension(self):  # CONSIDER  beautify this crud!
        return 0

    def validate_predicate(self):
        return  # looks good! (-:

    def enforce(self, condition, diagnostic):
        if not condition:
            msg = self.format_fault(diagnostic)
            raise SyntaxError(msg)

    def format_fault(self, diagnostic):
        parent_reconstruction = ''
        if self.parent:
            parent_reconstruction = self.parent.reconstruction().replace('\n', '\\n')
        reconstruction = self.reconstruction().replace('\n', '\\n')
        args = (self.get_filename(), self.line_number, parent_reconstruction, reconstruction, diagnostic)
        args = tuple([to_unicode(i) for i in args])
        return u'\n  File "%s", line %s, in %s\n    %s\n%s' % args

    def reconstruction(self):
        predicate = self.predicate
        try:
            predicate = predicate.decode('utf-8')
        except (UnicodeDecodeError, UnicodeEncodeError, AttributeError):
            pass
        recon = u'%s%s: %s' % (self.prefix(), self.concept, predicate)
        if recon[-1] != u'\n':
            recon += u'\n'
        return recon

    def get_filename(self):
        node = self

        while node:
            if not node.parent and hasattr(node, 'filename'):
                return node.filename
            node = node.parent

        return None


class Parser:
    def __init__(self, language=None):
        self.thangs = [
            Feature, Scenario,
            Given, When, Then, And, But,
            Row, Comment, Step
        ]
        self.steps = []
        self.language = DEFAULT_LANGUAGE if language is None else language

    def parse_file(self, filename):
        prose = open(filename, 'r').read()
        try:
            prose = prose.decode('utf-8')
        except (UnicodeEncodeError, AttributeError):
            pass
        self.parse_features(prose)
        self.steps[0].filename = filename
        return self

    def parse_features(self, prose):
        self.parse_feature(prose)
        return self

    def _create_matcher(self, suite, matcher_classes):
        root_matcher = None
        for matcher_class in matcher_classes:
            matcher = matcher_class(suite)
            if root_matcher is not None:
                root_matcher.add_matcher(matcher)
            else:
                root_matcher = matcher
        return root_matcher

    def evaluate(self, suite, matchers=None):
        if matchers is None:
            matchers = [RegexpStepMatcher, ParseStepMatcher, MethodNameStepMatcher]
        matcher = self._create_matcher(suite, matchers)
        test_visitor = TestVisitor(suite, matcher)
        step_matcher_visitor = StepMatcherVisitor(suite, matcher)
        try:
            self.rip(test_visitor)
        except MissingStepError:
            self.rip(step_matcher_visitor)

    def report(self, suite):
        rv = ReportVisitor(suite)
        self.rip(rv)
        return rv

    def rip(self, visitor):
        if self.steps != []:
            node = self.steps[0]
            visitor.before_feature(node)
            try:
                node.evaluate_steps(visitor)
            finally:
                visitor.after_feature(node)

    def parse_language_directive(self, line):
        """ Parse language directive.

        :param str line: line to parse
        :returns: True if line contains correct language directive
        :side effects: sets self.language to parsed language
        """
        match = LANGUAGE_RE.match(line)
        if match:
            self.language = match.groups()[0]
            return True
        return False

    def parse_feature(self, lines):
        lines = to_unicode(lines)
        self.line_number = 0

        for self.line in lines.split(u'\n'):
            self.line_number += 1

            if self.parse_language_directive(self.line):
                continue

            if not self.line or self.line.startswith('#'):
                continue

            if not self.anneal_last_broken_line() and not self._parse_line():
                if 0 < len(self.steps):
                    self._append_to_previous_node()
                else:
                    s = Step()
                    s.concept = '???'
                    s.predicate = self.line
                    s.line_number = self.line_number
                    feature_name = TRANSLATIONS[self.language].get('feature', u'Feature')
                    feature_name = feature_name.replace('|', ' or ')
                    s.enforce(False, u'feature files must start with a %s' % feature_name)

        return self.steps

    def anneal_last_broken_line(self):
        if self.steps == []:
            return False  # CONSIDER  no need me
        last_line = self.last_node.predicate

        if re.search(r'\\\s*$', last_line):
            last = self.last_node
            last.predicate += '\n' + self.line
            return True

        return False

#  TODO  permit line breakers in comments
#    | Given a table with one row
#        \| i \| be \| a \| lonely \| row |  table with only one row, line 1

    def _parse_line(self):
        self.line = self.line.rstrip()

        for klass in self.thangs:
            self.thang = klass(language=self.language)
            rx = self.thang._my_regex()
            m = re.compile(rx).match(self.line)

            if m and len(m.groups()) > 0:
                return self._register_line(m.groups())

    def _register_line(self, groups):
        predicate = ''
        if len(groups) > 1:
            predicate = groups[1]
        node = self.thang
        node._parse(predicate, self.steps, self.line_number)
        self.steps.append(node)
        self.last_node = node
        return node

    def _append_to_previous_node(self):
        previous = self.steps[-1]
        previous.predicate += '\n' + self.line.strip()
        previous.predicate = previous.predicate.strip()
        previous.validate_predicate()


class Feature(Morelia):

    def __str__(self):
        return u'Feature'

    def my_parent_type(self):
        return None

    def test_step(self, suite, matcher):
        self.enforce(0 < len(self.steps), 'Feature without Scenario(s)')

    def to_html(self):
        return ['''\n<div><table>
                <tr style="background-color: #aaffbb;" width="100%%">
                <td align="right" valign="top" width="100"><em>%s</em>:</td>
                <td colspan="101">%s</td>
                </tr></table></div>''' % (self.concept, _clean_html(self.predicate)), '']


class Scenario(Morelia):

    def __str__(self):
        return u'Scenario'

    def my_parent_type(self):
        return Feature

    def evaluate_steps(self, visitor):
        step_schedule = visitor.step_schedule(self)  # TODO  test this permuter directly (and rename it already)

        for step_indices in step_schedule:   # TODO  think of a way to TDD this C-:
            schedule = visitor.permute_schedule(self)

            for indices in schedule:
                self.row_indices = indices
                self.evaluate_test_case(visitor, step_indices)  # note this works on reports too!

    def evaluate_test_case(self, visitor, step_indices):  # note this permutes reports too!
        self.enforce(0 < len(self.steps), 'Scenario without step(s) - Step, Given, When, Then, And, or #')

        visitor.before_scenario(self)
        try:
            self.result = visitor.visit(self)

            for idx, step in enumerate(self.steps):
                if idx in step_indices:
                    step.evaluate_steps(visitor)
        finally:
            visitor.after_scenario(self)

    def permute_schedule(self):  # TODO  rename to permute_row_schedule
        dims = self.count_Row_dimensions()
        return _permute_indices(dims)

    def step_schedule(self):  # TODO  rename to permute_step_schedule !
        sched = []
        pre_slug = []

        #  TODO   deal with steps w/o whens

        for idx, s in enumerate(self.steps):
            if s.__class__ == When:
                break
            else:
                pre_slug.append(idx)

        for idx, s in enumerate(self.steps):
            if s.__class__ == When:
                slug = pre_slug[:]
                slug.append(idx)

                for idx in range(idx + 1, len(self.steps)):
                    s = self.steps[idx]
                    if s.__class__ == When:
                        break
                    slug.append(idx)

                sched.append(slug)

        if sched == []:
            return [pre_slug]
        return sched

    def count_Row_dimensions(self):
        return [step.count_dimensions() for step in self.steps]

    def reconstruction(self):
        return '\n' + self.concept + ': ' + self.predicate

    def to_html(self):
        return ['''\n<div><table width="100%%">
                <tr style="background-color: #cdffb8;">
                <td align="right" valign="top" width="100"><em>%s</em>:</td>
                <td colspan="101">%s</td></tr>''' % (self.concept, _clean_html(self.predicate)),
                '</table></div>']


class Step(Morelia):

    def prefix(self):
        return '  '

    def __str__(self):
        return u'Step'

    def my_parent_type(self):
        return Scenario

    def find_step(self, suite, matcher):
        predicate = self.predicate
        augmented_predicate = self._augment_predicate()
        method, args, kwargs = matcher.find(predicate, augmented_predicate)
        if method:
            return method, args, kwargs

        suggest = matcher.suggest(predicate)
        raise MissingStepError(predicate, suggest)

    def _augment_predicate(self):  # CONSIDER  unsucktacularize me pleeeeeeze
        if self.parent is None:
            return self.predicate
        dims = self.parent.count_Row_dimensions()
        if set(dims) == set([0]):
            return self.predicate
        rep = re.compile(r'\<(\w+)\>')
        replitrons = rep.findall(self.predicate)
        if replitrons == []:
            return self.predicate
        self.copy = self.predicate[:]
        row_indices = self.parent.row_indices

        for self.replitron in replitrons:
            for x in range(0, len(row_indices)):
                self.table = self.parent.steps[x].steps

                if self.table != []:
                    q = 0

                    for self.title in self.table[0].harvest():
                        self.replace_replitron(x, q, row_indices)
                        q += 1

        return self.copy

    def evaluate(self, suite, matcher):
        method, args, kwargs = self.find_step(suite, matcher)
        method(*args, **kwargs)

    def test_step(self, suite, matcher):
        try:
            self.evaluate(suite, matcher)
        except (Exception, SyntaxError) as e:
            new_exception = self.format_fault(to_unicode(e))
            e.args = (new_exception,) + (e.args[1:])
            if type(e) == SyntaxError:
                raise SyntaxError(repr(new_exception))
            raise

    def replace_replitron(self, x, q, row_indices):
        if self.title != self.replitron:
            return
        at = row_indices[x] + 1

        if at >= len(self.table):
            print('CONSIDER this should never happen')
            return

        #  CONSIDER  we hit this too many times - hit once and stash the result
        #  CONSIDER  better diagnostics when we miss these

        stick = self.table[at].harvest()
        found = stick[q]  # CONSIDER  this array overrun is what you get when your table is ragged
        # CONSIDER  only if it's not nothing?
        found = found.replace('\n', '\\n')  # CONSIDER  crack the multi-line argument bug, and take this hack out!
        self.copy = self.copy.replace('<%s>' % self.replitron, found)

        # CONSIDER  mix replitrons and matchers!

    def to_html(self):
        return '\n<tr><td align="right" valign="top"><em>' + self.concept + '</em></td><td colspan="101">' + _clean_html(self.predicate) + '</td></tr>', ''


class Given(Step):  # CONSIDER  distinguish these by fault signatures!

    def __str__(self):
        return u'Given'

    def prefix(self):
        return '  '


class When(Step):  # TODO  cycle these against the Scenario

    def __str__(self):
        return u'When'

    def prefix(self):
        return '   '

    def to_html(self):
        return '\n<tr style="background-color: #cdffb8; background: url(http://www.zeroplayer.com/images/stuff/aqua_gradient.png) no-repeat; background-size: 100%;"><td align="right" valign="top"><em>' + self.concept + '</em></td><td colspan="101">' + _clean_html(self.predicate) + '</td></tr>', ''


class Then(Step):

    def __str__(self):
        return u'Then'

    def prefix(self):
        return '   '


class And(Step):

    def __str__(self):
        return u'And'

    def prefix(self):
        return '    '


class But(And):

    def __str__(self):
        return u'But'

#  CONSIDER  how to validate that every row you think you wrote actually ran?


class Row(Morelia):

    def __str__(self):
        return u'Row'

    def i_look_like(self):
        return r'\|'

    def my_parent_type(self):
        return Step

    def prefix(self):
        return '        '

    def reconstruction(self):  # TODO  strip the reconstruction at error time
        recon = self.prefix() + '| ' + self.predicate
        if recon[-1] != '\n':
            recon += '\n'
        return recon

    def to_html(self):
        html = '\n<tr><td></td>'
        idx = self.parent.steps.index(self)
        em = 'span'
        if idx == 0:
            color = 'silver'
            em = 'em'
        elif ((2 + idx) / 3) % 2 == 0:
            color = '#eeffff'
        else:
            color = '#ffffee'

        for col in self.harvest():
            html += '<td style="background-color: %s;"><%s>' % (color, em) + _clean_html(col) + '</%s></td>' % em

        html += '<td>&#160;</td></tr>'  # CONSIDER  the table needn't stretch out so!
        return html, ''

    def count_dimension(self):
        if self is self.parent.steps[0]:
            # header row
            return 0
        return 1  # TODO  raise an error (if the table has one row!)

    def harvest(self):
        row = re.split(r' \|', re.sub(r'\|$', '', self.predicate))
        row = [s.strip() for s in row]
        return row

#  TODO  sample data with "post-it haiku"
#  CONSIDER  trailing comments


class Comment(Morelia):

    def __str__(self):
        return u'Comment'

    def i_look_like(self):
        return r'\#'

    def my_parent_type(self):
        return Morelia  # aka "any"

    def _my_regex(self):
        name = self.i_look_like()
        return r'\s*(' + name + ')(.*)'

    def validate_predicate(self):
        self.enforce(self.predicate.count('\n') == 0, 'linefeed in comment')

    def reconstruction(self):
        recon = '  # ' + self.predicate
        if recon[-1] != '\n':
            recon += '\n'
        return recon

    def to_html(self):
        return '\n# <em>' + _clean_html(self.predicate) + '</em><br/>', ''


def _special_range(n):  # CONSIDER  better name
    return moves.range(n) if n else [0]


def _permute_indices(arr):
    product_args = list(_imap(arr))
    result = list(itertools.product(*product_args))
    return result
    #  tx to Chris Rebert, et al, on the Python newsgroup for curing my brainlock here!!


def _imap(*iterables):
    iterables = [iter(i) for i in iterables]
    while True:
        args = [next(i) for i in iterables]
        yield _special_range(*args)


def _clean_html(string):
    return string.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;').replace("'", '&#39;')

#  CONSIDER  display all missing steps not just the first
#  ERGO  Morelia should raise a form in any state!
#  ERGO  get Morelia working with more Pythons - virtualenv it!
#  ERGO  moralia should try the regex first then the step name
#  ERGO  pay for "Bartender" by Sacred Hoop


if __name__ == '__main__':
    import os
    os.system('python ../tests/morelia_suite.py')   # NOTE  this might not return the correct shell value



import unittest
from collections import OrderedDict
from mock import Mock, sentinel

from morelia.exceptions import MissingStepError
from morelia.visitors import StepMatcherVisitor


class StepMatcherVisitorVisitTestCase(unittest.TestCase):
    """ Test :py:meth:`StepMatcherVisitor.visit`. """

    def test_should_find_step(self):
        """ Scenario: step found """
        # Arrange
        obj = StepMatcherVisitor(sentinel.suite, sentinel.matcher)
        node = Mock()
        # Act
        obj.visit(node)
        # Assert
        node.find_step.assert_called_once_with(sentinel.suite, sentinel.matcher)

    def test_should_not_find_step(self):
        """ Scenario: step not found """
        # Arrange
        obj = StepMatcherVisitor(sentinel.suite, sentinel.matcher)
        node = Mock()
        suggest = 'suggest'
        node.find_step.side_effect = [MissingStepError('predicate', suggest)]
        # Act
        obj.visit(node)
        # Assert
        node.find_step.assert_called_once_with(sentinel.suite, sentinel.matcher)
        self.assertTrue(suggest in obj._not_matched)


class StepMatcherVisitorAfterFeatureTestCase(unittest.TestCase):
    """ Test :py:meth:`StepMatcherVisitor.after_feature`. """

    def test_should_fail_with_suggest(self):
        """ Scenario: fail with suggset """
        # Arrange
        suite = Mock()
        obj = StepMatcherVisitor(suite, sentinel.matcher)
        obj._not_matched = OrderedDict([('suggest string', True)])
        # Act
        obj.after_feature(sentinel.node)
        # Assert
        suite.fail.assert_called_once_with(u'Cannot match steps:\n\nsuggest string')

    def test_should_not_fail(self):
        """ Scenario: no steps missing """
        # Arrange
        suite = Mock()
        obj = StepMatcherVisitor(suite, sentinel.matcher)
        obj._not_matched = OrderedDict()
        # Act
        obj.after_feature(sentinel.node)
        # Assert
        self.assertFalse(suite.fail.called)
